/*
 * Copyright (C) 2009 by Pablo Alvarado
 * 
 * This file is part of the lecture CE-5201 Digital Image Processing and
 * Analysis, at the Costa Rica Institute of Technology.
 *
 * The LTI-Lib-2 is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the authors nor the names of its contributors may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/** 
 * \file   asm.cpp
 *         Examples of ASM use
 * \author Pablo Alvarado
 * \date   15.08.2009
 * revisions ..: $Id: asm.cpp,v 1.10 2010/06/08 15:32:07 alvarado Exp $
 */

// LTI-Lib Headers
#include <ltiObject.h>
#include <ltiMath.h>     // General lti:: math and <cmath> functionality
#include <ltiTimer.h>    // To measure time

#include <ltiLispStreamHandler.h>

#include <ltiActiveShapeModel.h>
#include <ltiPDMGeneration.h>

#include <ltiDraw.h>
#include <ltiViewer2D.h>
#include <ltiImage.h>
#include <ltiColors.h>
#include <ltiPolygonPoints.h>

// Standard Headers: from ANSI C and GNU C Library
#include <cstdlib>  // Standard Library for C++
#include <getopt.h> // Functions to parse the command line arguments

// Standard Headers: STL
#include <iostream>
#include <string>
#include <fstream>

// Mosaic files
#include "asm.h"

// Debug

// Ensure that the STL streaming is used.
using std::cout;
using std::cerr;
using std::endl;

#undef _LTI_DEBUG
//#define _LTI_DEBUG 4
#include "ltiDebug.h"


ASM::ASM(int argc, char* argv[]) 
  : task_(Exit),line_(true) {
  parse(argc,argv);

  // Create a triangle (equilateral)
  triangleIso_.allocate(12);
  int i=0;
  triangleIso_.at(i++).set(1.00f,2.00f);
  triangleIso_.at(i++).set(1.25f,1.50f);
  triangleIso_.at(i++).set(1.50f,1.00f);
  triangleIso_.at(i++).set(1.75f,0.50f);
  triangleIso_.at(i++).set(2.00f,0.00f);

  triangleIso_.at(i++).set(1.50f,0.00f);
  triangleIso_.at(i++).set(1.00f,0.00f);
  triangleIso_.at(i++).set(0.50f,0.00f);
  triangleIso_.at(i++).set(0.00f,0.00f);

  triangleIso_.at(i++).set(0.25f,0.50f);
  triangleIso_.at(i++).set(0.50f,1.00f);
  triangleIso_.at(i++).set(0.75f,1.50f);


  triangleRect_.allocate(12);
  i=0;

  // Create a triangle (rectangle)
  triangleRect_.at(i++).set(0.00f,2.00f);
  triangleRect_.at(i++).set(0.50f,1.50f);
  triangleRect_.at(i++).set(1.00f,1.00f);
  triangleRect_.at(i++).set(1.50f,0.50f);
  triangleRect_.at(i++).set(2.00f,0.00f);

  triangleRect_.at(i++).set(1.50f,0.00f);
  triangleRect_.at(i++).set(1.00f,0.00f);
  triangleRect_.at(i++).set(0.50f,0.00f);
  triangleRect_.at(i++).set(0.00f,0.00f);

  triangleRect_.at(i++).set(0.00f,0.50f);
  triangleRect_.at(i++).set(0.00f,1.00f);
  triangleRect_.at(i++).set(0.00f,1.50f);

  square_.allocate(12);
  i=0;

  // Square
  square_.at(i++).set(0.00f      ,2.00f);
  square_.at(i++).set(2.00f/3.00f,2.00f);
  square_.at(i++).set(4.00f/3.00f,2.00f);
  square_.at(i++).set(2.00f      ,2.00f);

  square_.at(i++).set(2.00f      ,4.00f/3.00f);
  square_.at(i++).set(2.00f      ,2.00f/3.00f);
  square_.at(i++).set(2.00f      ,0.00f);

  square_.at(i++).set(4.00f/3.00f,0.00f);
  square_.at(i++).set(2.00f/3.00f,0.00f);
  square_.at(i++).set(0.00f      ,0.00f);

  square_.at(i++).set(0.00f      ,2.00f/3.00f);
  square_.at(i++).set(0.00f      ,4.00f/3.00f);

  circle_.allocate(12);

  // Circle
  for (int a=0;a<12; ++a) {
    circle_.at(a).set(sin(a*lti::Pi/6)+1.00f,cos(a*lti::Pi/6)+1.00f);
  }

}


/*
 * Help 
 */
void ASM::usage() const {
  cout <<
    "usage: asm [options] \n\n" \
    "       -c <file>        create an example model and saves it in the\n" \
    "                        given file\n" \
    "       --create <dir>   like -c\n" \
    "       -v <file>        view the model stored in the given file.\n" \
    "                        Use the keys 1-9 to select which eigenvalue is\n"
    "                        modified, and the arrows to increase or decrease"\
    "                        it\n"
    "       --view <file>    like -v\n" \
    "       -s <file>        read file of shapes to be fit with the model\n"\
    "       --shapes <file>  like -s\n"\
    "       -h | --help      show this message.\n"\
    "       <file>           input image" << std::endl;    
}

void ASM::help() const {
  cout << 
    " 0-9      Select which dimension to change.\n" \
    " a-k      Get the coefficients for the training shapes.\n" \
    " l        Toggle lines painting of the shapes.\n" \
    " m        Reset mean shape and shape counter.\n" \
    " n        Next shape.\n" \
    " p        Previous shape.\n" \
    " Arrows   Increase/Decrease dimension.\n" \
    " ?        Print this message.\n" << std::endl;
}


/*
 * Read shapes from file
 */
bool ASM::readShapes(const std::string file) {
  std::ifstream in(file.c_str());
  if (in) {
    lti::lispStreamHandler lsh(in);
    bool ok=true;
    int counter=0;
    lti::pointDistributionModel::shape s;
    do {
      ok=s.read(lsh);
      if (ok) {
        shapes_.push_back(s);
        counter++;
      }
    } while(ok);
    in.close();
    std::cout << counter << " shapes read from '" << file << "'" << std::endl;
    return !shapes_.empty();
  }

  return false;
}

/*
 * Parse the line command arguments
 */
void ASM::parse(int argc, char*argv[]) {

  int c;

  // We use the standard getopt.h functions here to parse the arguments.
  // Check the documentation of getopt.h for more information on this
  
  // structure for the long options. 
  static struct option lopts[] = {
    {"create",required_argument,0,'c'},
    {"view",required_argument,0,'b'},
    {"shapes",required_argument,0,'s'},
    {"help",no_argument,0,'h'},
    {0,0,0,0}
  };

  int optionIdx;

  while ((c = getopt_long(argc, argv, "c:v:s:h", lopts,&optionIdx)) != -1) {
    switch (c) {
    case 'c':
      task_=Create;
      fileASM_=optarg; // get the argument provided to -c
      break;
    case 'v':
      task_=View;
      fileASM_=optarg; // get the argument provided to -v
      break;
    case 's':
      readShapes(optarg);
      break;
    case 'h':
      break;
    default:
      cerr << "Option '-" << static_cast<char>(c) << "' not recognized." 
           << endl;
    }
  }

}

bool ASM::show(const lti::pointDistributionModel::shape& shape) const {
 
  lti::pointDistributionModel::shape::const_iterator it;
  
  lti::ipolygonPoints poly;
  lti::draw<lti::rgbaPixel> painter;
  lti::image canvas(256,256,lti::White);
  painter.use(canvas);

  const int dim = lti::min(canvas.rows(),canvas.columns());
  const lti::ipoint offset(dim/4,dim/4);
  
  poly.clear();
    
  // paint the shape
  for (it=shape.begin();it!=shape.end();++it) {
    poly.push_back(lti::ipoint((*it).x*dim/4,(*it).y*dim/4));
  }

  painter.setColor(lti::Blue);
  painter.polygon(poly,false,0.0f,offset);
  
  painter.setColor(lti::Red);
  painter.set(poly,offset,false);


  static lti::viewer2D view;
  lti::viewer2D::interaction action;
  view.show(canvas);
  lti::ipoint pos;
  
  do {
    view.waitInteraction(action,pos);
  } while (action!=lti::viewer2D::KeyPressed);

  return true;
}

ASM::eTasks ASM::checkTask() const {
  return task_;
}

bool ASM::createASM() {

  lti::pdmGeneration pdm;

  if (shapes_.empty()) {

    pdm.consider(triangleIso_);
    
#ifdef _LTI_DEBUG
    show(triangleIso_);
#endif
    
    pdm.consider(triangleRect_);
    
#ifdef _LTI_DEBUG
    show(triangleRect_);
#endif
    
    pdm.consider(square_);
    
#ifdef _LTI_DEBUG
    show(square_);
#endif
    
    pdm.consider(circle_);
    
#ifdef _LTI_DEBUG
    show(circle_);
#endif
  } else {
    for (unsigned int i=0;i<shapes_.size(); ++i) {
      pdm.consider(shapes_[i]);
#ifdef _LTI_DEBUG
      show(shapes_[i]);
#endif
    }
  }

  lti::pointDistributionModel model;

  if (pdm.apply(model)) {

    std::cout << "Model successfully created. Saving...";
    std::cout.flush();

    lti::lispStreamHandler lsh;
    std::ofstream out(fileASM_.c_str());
    lsh.use(out);
    if (model.write(lsh)) {
      std::cout << "done." << std::endl;
    } else {
      std::cout << "failed." << std::endl;
    }
    out << std::endl;
    out.close();

    return true;
  } else {
    std::cout << "Model could not be created." << std::endl;
  }

  return false;
}

bool ASM::viewASM() {
  lti::activeShapeModel::parameters asmPar;

  lti::lispStreamHandler lsh;
  std::ifstream in(fileASM_.c_str());
  lsh.use(in);
  if (!asmPar.model.read(lsh)) {
    std::cerr << "Error reading " << fileASM_ << ": " 
              << lsh.getStatusString() << std::endl;

    return false;
  }

  help();

  if (asmPar.model.meanShape.empty()) {
    std::cerr<<"Model in '" << fileASM_ << "' has no mean shape" << std::endl;
    return false;
  }

  // Analize mean shape for bounding box
  lti::ipoint mins(asmPar.model.meanShape.at(0));
  lti::ipoint maxs(asmPar.model.meanShape.at(0));

  // find the bounding box
  for (int i=1;i<asmPar.model.meanShape.size();++i) {
    lti::minmax(lti::iround(asmPar.model.meanShape.at(i).x),mins.x,maxs.x);
    lti::minmax(lti::iround(asmPar.model.meanShape.at(i).y),mins.y,maxs.y);
  }

  const int border = 128;
  lti::ipoint delta = maxs-mins;
  float f=lti::max(delta.x,delta.y);
  float dim=256.0f/f;

  const int rows = delta.y*dim + 1 + 2*border;
  const int cols = delta.x*dim + 1 + 2*border;
  
  lti::activeShapeModel theAsm(asmPar);

  lti::image canvas(rows,cols);
  lti::viewer2D::interaction action;
  lti::viewer2D view("ASM");
  lti::ipoint pos;

  lti::dvector weights(asmPar.model.varianceVector.size(),0.0);
  lti::pointDistributionModel::shape shape;
  lti::pointDistributionModel::shape* shapeTrained = 0;
  lti::pointDistributionModel::shape::const_iterator it;
  
  lti::ipolygonPoints poly;

  lti::draw<lti::rgbaPixel> painter;
  painter.use(canvas);
  const lti::ipoint offset = lti::ipoint(border,border)-mins*dim;
  
  int currentDim = 0;
  
  if (shapes_.empty()) {
    shapes_.push_back(triangleIso_);
    shapes_.push_back(triangleRect_);
    shapes_.push_back(circle_);
    shapes_.push_back(square_);
  }

  lti::matrixTransform< float >::parameters mtp;
  int shapeCounter = 0;
  lti::pointDistributionModel::shape sh;

  do {
    canvas.fill(lti::White);

    theAsm.convertWeightsVectorToShape(weights,shape);

    // user wants to see the trained shape
    if (shapeTrained != 0) {
      poly.clear();
      
      // paint the shape
      for (it=shapeTrained->begin();it!=shapeTrained->end();++it) {
        poly.push_back(lti::ipoint((*it).x*dim,(*it).y*dim));
      }
      
      if (line_) {
        painter.setColor(lti::Yellow);
        painter.polygon(poly,false,0.0f,offset);
      }
      
      painter.setColor(lti::Green);
      painter.set(poly,offset,false);
    }


    poly.clear();
    
    // paint the shape
    for (it=shape.begin();it!=shape.end();++it) {
      poly.push_back(lti::ipoint((*it).x*dim,(*it).y*dim));
    }

    if (line_) {
      painter.setColor(lti::Blue);
      painter.polygon(poly,false,0.0f,offset);
    }

    painter.setColor(lti::Red);
    painter.set(poly,offset,false);

    view.show(canvas);
    view.waitInteraction(action,pos);

    switch (action.action) {
    case lti::viewer2D::KeyPressed:
      switch(action.key) {
      case '0': 
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
        currentDim=action.key-'0';
        std::cout << "\nModifying " << currentDim << " dimension." << std::endl;
        break;
      case 'a':
      case 'b':
      case 'c':
      case 'd':
      case 'e':
      case 'f':
      case 'g':
      case 'h':
      case 'i':
      case 'j':
      case 'k': {
        shapeCounter = lti::within(action.key-'a',0,int(shapes_.size())-1);
        sh.copy(shapes_[shapeCounter]);
        theAsm.alignShape(asmPar.model.meanShape,sh,mtp);
        lti::dvector vct;
        theAsm.convertShapeToWeightsVector(sh,vct);
        shapeTrained = &sh;
        weights.copy(vct);
        std::cout << "\nShape[" << shapeCounter << "]: " << weights 
                  << std::endl;
      } break;     
      case 'l':
        line_=!line_;
        break;
      case 'p': {
        shapeCounter--;
        if (shapeCounter<0) {
          shapeCounter=shapes_.size()-1;
        }
        sh.copy(shapes_[shapeCounter]);
        theAsm.alignShape(asmPar.model.meanShape,sh,mtp);
        lti::dvector vct;
        theAsm.convertShapeToWeightsVector(sh,vct);
        shapeTrained = &sh;
        weights.copy(vct);
        std::cout << "\nShape[" << shapeCounter << "]: " << weights 
                  << std::endl;
      } break;
      case 'n': {
        shapeCounter++;
        if (shapeCounter>=int(shapes_.size())) {
          shapeCounter=0;
        }
        sh.copy(shapes_[shapeCounter]);
        theAsm.alignShape(asmPar.model.meanShape,sh,mtp);
        lti::dvector vct;
        theAsm.convertShapeToWeightsVector(sh,vct);
        shapeTrained = &sh;
        weights.copy(vct);
        std::cout << "\nShape[" << shapeCounter << "]: " << weights 
                  << std::endl;
      } break;
      case 'm':
        weights.fill(0.0);
        shapeTrained=0;
        std::cout << "\nMean shape set" << std::endl;
        break;
      case '?':
        help();
        break;
      case 65362: // up
        if (currentDim<weights.size()) {
          weights.at(currentDim)+=
            sqrt(asmPar.model.varianceVector.at(currentDim));
          std::cout << "\r  Weight[" << currentDim << "] = "
                   << weights.at(currentDim) << "               " ;
          std::cout.flush();
        }
        break;
      case 65363: // right
        if (currentDim<weights.size()) {
          weights.at(currentDim)+=
            sqrt(asmPar.model.varianceVector.at(currentDim))/10;
          std::cout << "\r  Weight[" << currentDim << "] = "
                    << weights.at(currentDim) << "               ";
          std::cout.flush();
        }
        break;
      case 65361: // left
        if (currentDim<weights.size()) {
          weights.at(currentDim)-=
            sqrt(asmPar.model.varianceVector.at(currentDim))/10;
          std::cout << "\r  Weight[" << currentDim << "] = "
                    << weights.at(currentDim) << "               ";
          std::cout.flush();
        }
        break;        
      case 65364: // down
        if (currentDim<weights.size()) {
          weights.at(currentDim)-=
            sqrt(asmPar.model.varianceVector.at(currentDim));
          std::cout << "\r  Weight[" << currentDim << "] = "
                    << weights.at(currentDim) << "               ";
          std::cout.flush();
        }
        break;        
      default:
        std::cout << "\nKey " << action.key << " unassigned" << std::endl;
        break;
      }
      break;
    default:
      break;
    }

  } while (action.action != lti::viewer2D::Closed);

  return false;
}

bool ASM::apply() {
  switch(checkTask()) {
  case Create:
    return createASM();
    break;
  case View:
    return viewASM();
    break;
  case Exit:
    usage();
    return true;
    break;
  default:
    usage();
  }
  return true;
}

/*
 * Main method
 */
int main(int argc, char* argv[]) {
  
  ASM me(argc,argv);
  if (me.apply()) {
    return EXIT_SUCCESS;
  }

  return EXIT_FAILURE;
    

}
