/*
 * Copyright (C) 2008 by Pablo Alvarado
 * 
 * This file is part of the LTI-Computer Vision Library 2 (LTI-Lib-2)
 *
 * The LTI-Lib-2 is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the authors nor the names of its contributors may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

// LTI-Lib Headers
#include "ltiObject.h"
#include "ltiRbf.h"
#include "ltiMlp.h"
#include "ltiCombination.h"

#include "ltiLispStreamHandler.h"
#include "ltiClassifier2DVisualizer.h"
#include "ltiViewer2D.h"

#include "ltiStreamProgressInfo.h"


// Standard Headers
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <vector>
#include <list>

using std::cout;
using std::cerr;
using std::endl;

namespace lti {

  /**
   * A pseudo classifier, that combines the results of all given classifiers
   */ 
  class combinedClassifier : public supervisedInstanceClassifier {
  public:
    /**
     * The list of classifiers, whose output should be combined
     */ 
    std::list<supervisedInstanceClassifier*> classifiers;

    /**
     * The combiner used
     */
    combination combiner;
    
    /**
     * Constructor
     */
    combinedClassifier() {
    }

    /**
     * Destructor
     */
    virtual ~combinedClassifier() {
    }

    /**
     * One of the virtual "must have" methods: clone
     */
    combinedClassifier* clone() const {
      return new combinedClassifier(*this);
    }
    
    /**
     * One of the virtual "must have" methods: newInstance
     */
    combinedClassifier* newInstance() const {
      return new combinedClassifier;
    }

    /**
     * One of the virtual "must have" methods: name
     */
    virtual const std::string& name() const {
      _LTI_RETURN_CLASS_NAME
    }

    /**
     * Must have a train() method, even if it does not do anything
     */
    virtual bool train(const dmatrix& input,
                       const ivector& ids) {
      return false; // not implemented, just for classifying
    }

    /**
     * The classification method.
     *
     * This can be done much more efficiently, without the unneceesary
     * creation each time of the partial results, but for the demonstration
     * purposes, this should suffice.
     */
    virtual bool classify(const dvector& feature, result& res) const {
      std::list<classifier::result> all;
      classifier::result tmp;
      std::list<supervisedInstanceClassifier*>::const_iterator it;
      for (it=classifiers.begin();it!=classifiers.end();++it) {
        all.push_back(tmp);
        (*it)->classify(feature,all.back());
      }
      return combiner.apply(all,res);
    }    
  };


  /**
   * The real class
   */
  class tester {
  public:
    void operator()(int argc,char *argv[]);
  };

  void tester::operator()(int argc,char *argv[]) {

    combination::parameters combiPar;

    // read the configuration
    // Radial Basis Functions classifier (RBF)
    rbf::parameters rbfPar;
    rbfPar.sigmaFactor = 0.7;
    rbfPar.nbHiddenNeurons = 5;

    // Multi-Layer Perceptron classifier (MLP)
    mlp::parameters mlpPar;
    mlpPar.hiddenUnits.assign(2,6); // 2 hidden layers with 6 units each
    mlpPar.activationFunctions.assign(mlpPar.hiddenUnits.size()+1,
                                     "sigmoidFunctor");

    mlpPar.trainingMode = lti::mlp::SteepestDescent;
    mlpPar.learnrate = 0.2;
    mlpPar.maxNumberOfEpochs=5000;

    bool readNet = false;

    // read configuration file
    static const char* cfile = "config.lsp";    
    std::ifstream in(cfile);
    lispStreamHandler lsh;
    
    if (in) {
      lsh.use(in);
      lti::read(lsh,"readNet",readNet);
      lti::read(lsh,"rbfPar",rbfPar);
      lti::read(lsh,"mlpPar",mlpPar);
      lti::read(lsh,"combiPar",combiPar);
      in.close();
    } else {
      std::cout << "Configuration file not found.  Saving " << cfile 
                << std::endl;
      std::ofstream out(cfile);
      lsh.use(out);
      lti::write(lsh,"readNet",readNet);
      lti::write(lsh,"rbfPar",rbfPar);
      lti::write(lsh,"mlpPar",mlpPar);
      lti::write(lsh,"combiPar",combiPar);

      out << std::endl;
      out.close();

      exit(0);
    }

    double inData[] = {    -1,-1,
                           -1, 0,
                           -1,+1,
                           +0,+1,
                           +1,+1,
                           +1,+0,
                           +1,-1,
                           +0,-1,
                           +0,+0};
    
    lti::dmatrix inputs(9,2,inData);          // training vectors
    
    int idsData[] = {1,0,1,0,1,0,1,0,1}; // and the respective ids
    lti::ivector ids(9,idsData);
    
    lti::rbf annRBF(rbfPar);  // our artificial neural networks
    lti::mlp annMLP(mlpPar);

    combinedClassifier combi; // the combiner pseudo-classifier
    combi.combiner.setParameters(combiPar);
        
    // we want to see some info while training
    streamProgressInfo prog(std::cout);
    annRBF.setProgressObject(prog);
    annMLP.setProgressObject(prog);
    
    
    if (!readNet) {
      std::cout << "TRAINING" << std::endl;

      // train the RBF network
      annRBF.train(inputs,ids);
     
      // let us save our network for future use
      // in the file called rbf.dat
      std::ofstream outRBF("rbf.dat");
      lsh.use(outRBF);
      
      // save the network
      annRBF.write(lsh);
      // close the file
      outRBF.close();

      // train the MLP network
      annMLP.train(inputs,ids);
     
      // let us save our network for future use
      // in the file called rbf.dat
      std::ofstream outMLP("mlp.dat");
      lsh.use(outMLP);
      
      // save the network
      annMLP.write(lsh);
      // close the file
      outMLP.close();

    } else {
      std::cout << "READING" << std::endl;
      std::ifstream inRBF("rbf.dat");
      lsh.use(inRBF);
      
      // save the network
      annRBF.read(lsh);

      std::ifstream inMLP("mlp.dat");
      lsh.use(inMLP);
      
      // save the network
      annMLP.read(lsh);

    }
    

    // show some results with the same training set:
    
    lti::rbf::result outv; // here we will get some classification results
    std::cout << std::endl << "Results: " << std::endl;
    
    int i,id;
    for (i=0;i<inputs.rows();++i) {
      annRBF.classify(inputs.getRow(i),outv);
      std::cout << "Input " << inputs.getRow(i) << " \tOutput RBF: ";
      outv.getId(outv.getWinner(),id);
      std::cout << id;
      if (id != ids.at(i)) {
        std::cout << " <- should be " << ids.at(i);
      }

      annMLP.classify(inputs.getRow(i),outv);
      std::cout << " \tOutput MLP: ";
      outv.getId(outv.getWinner(),id);
      std::cout << id;
      if (id != ids.at(i)) {
        std::cout << " <- should be " << ids.at(i);
      }
      
      std::cout << std::endl;
    }

    // view how the things were trained
    viewer2D viewRBF("RBF"),viewMLP("MLP"),viewCombi("Combi");
    classifier2DVisualizer::parameters vpar;
    vpar.lowerLeft.set(-2,-2);
    vpar.upperRight.set(2,2);
    vpar.imgSize.set(512,512);
    classifier2DVisualizer v(vpar);
    image canvas;
    v.apply(annRBF,inputs,canvas);
    viewRBF.show(canvas);

    v.apply(annMLP,inputs,canvas);
    viewMLP.show(canvas);

    combi.classifiers.clear();
    combi.classifiers.push_back(&annRBF);
    combi.classifiers.push_back(&annMLP);

    v.apply(combi,inputs,canvas);
    viewCombi.show(canvas);

    getchar();
  }

} // namespace lti

int main(int argc,char *argv[]) {
  try {
    lti::tester testObj;
    testObj(argc,argv);
    std::cout << "Test at the end!" << std::endl;
  }
  catch (lti::exception& exp) {
    std::cout << "An LTI::EXCEPTION was thrown: ";
    std::cout << exp.what() << std::endl;
  }
  catch (std::exception& exp) {
    std::cout << "std::exception was thrown: ";
    std::cout << exp.what() << std::endl;
  }
  catch (...) {
    std::cout << "Unknown exception thrown!" << std::endl;
  }
  return 0;
  
} // main
