/*
 * Copyright (C) 2008-2012
 * Pablo Alvarado
 *
 * This file is part of the LTI-Computer Vision Library 2 (LTI-Lib-2)
 *
 * The LTI-Lib-2 is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the authors nor the names of its contributors may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/** 
 * \file   colorProbability.cpp
 *         Contains an example for the use of the class
 *         lti::colorProbabilityMap.
 * \author Pablo Alvarado
 * \date   27.12.2008
 *
 * revisions ..: $Id: colorProbability.cpp,v 1.6 2012/08/06 03:04:57 alvarado Exp $
 */

#include "colorProbability.h"

#include <cstdlib>
#include <fstream>
#include <iostream>

#include <ltiLoadImageList.h>
#include <ltiIOImage.h>
#include <ltiViewer2D.h>
#include <ltiPassiveWait.h>

#include <ltiStatusMonitorCerr.h>

source::source(const eInput src,
               const std::list<std::string>& files) {
#ifdef _WITH_V4L2
  v4l2_ = 0;

  // read the parameters
  static const char* confFile = "v4l2.cfg";
  
  lti::v4l2::parameters param;
  
  // try to read the configuration file
  std::ifstream in(confFile);
  bool write=true;
  if (in) {
    lti::lispStreamHandler lsh;
    lsh.use(in);
    write=!param.read(lsh);
  }
  if (write) {
    // something went wrong reading, write a new configuration file
    std::ofstream out(confFile);
    lti::lispStreamHandler lsh;
    lsh.use(out);
    param.write(lsh);
    out << std::endl;
    out.close();
  }
  v4l2_ = new lti::v4l2(param);
#endif

  files_ = files;
  fileIterator_ = files_.begin();
  input_ = src;
}

source::~source() {
  delete v4l2_;
}

void source::dump() {
#ifdef _WITH_V4L2
  // read the parameters
  static const char* confFile = "v4l2.cfg";
  std::ofstream out(confFile);
  lti::lispStreamHandler lsh;
  lsh.use(out);

  std::cout << "Saving camera parameters" << std::endl;

  lti::v4l2::parameters param;
  v4l2_->dumpParameters(param);
  param.write(lsh);
  out << std::endl;
  out.close();
#endif
}

bool source::nextImage(lti::image& img) {
  std::string file;

  lti::ioImage loader;
  bool b=false;

  if (files_.empty()) {
    return false;
  }

  while (!b) {

    file = *fileIterator_;
    ++fileIterator_;
    if (fileIterator_ == files_.end()) {
      fileIterator_ = files_.begin();
    }

    std::cout << "Analyzing '" << file << "'" << std::endl;

    b = loader.load(file,img);
  };

  return b;
}

bool source::nextV4l2(lti::image& img) {
#ifdef _WITH_V4L2
  return v4l2_->apply(img);
#else
  std::cerr << "Support for V4l2 removed at compilation time." << std::endl;
  return false;
#endif
}


bool source::next(lti::image& img) {
  switch(input_) {
  case Files:
    return nextImage(img);
    break;
  case V4L2:
    return nextV4l2(img);
    break;
  default:
    return false;
  }
}



const char *const example::defaultConfigFile_ = "colorProbability.cfg";

example::example() {
  // initialization of the attributes
  configurationFile_  = defaultConfigFile_;
  objectModelFile_    = "objectModel.hist";
  nonObjectModelFile_ = "nonObjectModel.hist";
  cmePar_.maskIgnore.set(32,192);

  trainImages_        = "images.txt";
  maskSuffix_         = "_mask";

  sourceType_         = source::Files;

  trainMode_ = false;
  pause_ = false;


}

bool example::read(lti::ioHandler& handler) {
  bool b = true;

  b = lti::read(handler,"objectModelFile",objectModelFile_) && b;
  b = lti::read(handler,"nonObjectModelFile",nonObjectModelFile_) && b;
  b = lti::read(handler,"trainImages",trainImages_) && b;
  b = lti::read(handler,"maskSuffix",maskSuffix_) && b;
  b = lti::read(handler,"parameters",parameters_) && b;
  b = lti::read(handler,"estimation",cmePar_) && b;

  return b;
}

bool example::write(lti::ioHandler& handler) const {
  bool b = true;

  b = lti::write(handler,"objectModelFile",objectModelFile_) && b;
  b = lti::write(handler,"nonObjectModelFile",nonObjectModelFile_) && b;
  b = lti::write(handler,"trainImages",trainImages_) && b;
  b = lti::write(handler,"maskSuffix",maskSuffix_) && b;
  b = lti::write(handler,"parameters",parameters_) && b;
  b = lti::write(handler,"estimation",cmePar_) && b;

  return b;
}

void example::usage(int argc,char *argv[]) {
  std::cout << "\nUsage: " << argv[0] 
            << " [-t] [-c filename] [-h] [image1 [image2 ...] ]\n\n"; 
  std::cout << "  -t Train the models\n";
  std::cout << "  -c Use the given filename instead of the default '"
            << defaultConfigFile_ << "'\n";
  std::cout << "  -h Show this help\n";
  std::cout << "  -v Use the V4l2 class to get images from a camera\n";
  std::cout << "  -p Pause between each image (waits for viewer interaction"
            << std::endl;
}

bool example::parse(int argc,char *argv[],std::list<std::string>& files) {
  int i=1;

  files.clear();

  while (i<argc) {
    if (std::string(argv[i]) == "-c") {
      ++i;
      if (i<argc) {
        configurationFile_ = argv[i];
      } else {
        usage(argc,argv);
        return false; // report error
      }
    } else if (std::string(argv[i]) == "-t") {
      trainMode_ = true;
    } else if ( (std::string(argv[i]) == "-h") ||
                (std::string(argv[i]) == "--help") ) {
      usage(argc,argv);
      return false;
    } else if (std::string(argv[i]) == "-v") {
      sourceType_ = source::V4L2;
    } else if (std::string(argv[i]) == "-p") {
      pause_ = true;
    } else {
      files.push_back(argv[i]);
    }
    ++i;
  }

  return true;
}

int example::operator()(int argc,char *argv[]) {
  std::list<std::string> argFiles;

  if (!parse(argc,argv,argFiles)) {
    return EXIT_FAILURE;
  }
  
  //
  // read the configuration file, and write it if something went wrong
  //
  std::ifstream in(configurationFile_.c_str());
  bool writeConfig = true;
  if (in) {
    lti::lispStreamHandler lsh(in);
    writeConfig = !read(lsh);
  }

  if (writeConfig) {
    std::ofstream out(configurationFile_.c_str());
    lti::lispStreamHandler lsh(out);
    if (!write(lsh)) {
      std::cerr << "Could not write file '" << configurationFile_ << "'" 
                << std::endl;

      return EXIT_FAILURE;
    }
    out << std::endl;
    out.close();
  }

  in.close();

  //
  // Now the real job!
  //
  if (trainMode_) {
    train();
  } else {
    classify(argFiles);
  }

  return EXIT_SUCCESS;
}

std::string example::getMaskName(const std::string& imgName) const {
  std::string file(imgName);
  std::string::size_type pos = file.rfind('.');
  if (pos != std::string::npos) {
    file.insert(pos,maskSuffix_);
  } else {
    file+=maskSuffix_;
  }
  return file;
}



bool example::train() {

  std::cout << "Training color models" << std::endl;

  lti::colorModelEstimation objEstimator(cmePar_), nonObjEstimator(cmePar_);

  lti::ioImage maskLoader;
  lti::loadImageList loader;
  loader.useFileList(trainImages_);
  lti::image img;
  lti::channel8 mask;
  int n=0; // number of read images
  while (loader.hasNext()) {
    std::string file = loader.getNextFilename();
    std::cout << "  Reading '" << file << "'" << std::flush;
    if (loader.load(img)) {
      std::cout << " done." << std::endl;
      ++n;
      // try the mask
      std::string maskName = getMaskName(file);
      bool hasMask = ( maskLoader.load(maskName,mask) && 
                      (mask.size() == img.size()) );
     
      if (hasMask) {
        std::cout << "  Valid mask found." << std::endl;
        // if a valid mask exist, then consider object and non-object pixels
        objEstimator.consider(img,mask,nonObjEstimator);
      } else {
        // if no mask available, consider all pixels as part of the object.
        objEstimator.consider(img,mask);
      }
    } else {
      std::cout << " failed: " << loader.getStatusString() << std::endl;
    }
  }

  if (n>0) {
    objEstimator.apply(objectModel_);
    nonObjEstimator.apply(nonObjectModel_);

    // now save the models
    std::ofstream outObj(objectModelFile_.c_str());
    lti::lispStreamHandler lsh(outObj);
    if (!objectModel_.write(lsh)) {
      std::cerr << "Could not write '" <<
        objectModelFile_ << "': " << lsh.getStatusString() << std::endl;
    }
    outObj << std::endl;
    outObj.close();
    
    std::ofstream outNObj(nonObjectModelFile_.c_str());
    lsh.use(outNObj);
    if (!nonObjectModel_.write(lsh)) {
      std::cerr << "Could not write '" <<
        nonObjectModelFile_ << "': " << lsh.getStatusString() << std::endl;
    }
    outNObj << std::endl;
    outNObj.close();
    
    return true;
  } else {
    std::cerr << "No image loaded.  Train failed!" << std::endl;
  }
  
  return false;
}


bool example::classify(const std::list<std::string>& imgFiles) {
  // first, read both models
  std::ifstream inObj(objectModelFile_.c_str());
  lti::lispStreamHandler lsh(inObj);
  if (!objectModel_.read(lsh)) {
    std::cerr << "Could not read object model file '" << objectModelFile_
              << "'" << std::endl;
  }

  inObj.close();
    
  std::ifstream inNObj(nonObjectModelFile_.c_str());
  lsh.use(inNObj);
  if (!nonObjectModel_.read(lsh)) {
    std::cerr << "Could not read non-object model file '" << nonObjectModelFile_
              << "'" << std::endl;
  }
  inNObj.close();

  // second, initialize the color probability functor

  lti::colorProbabilityMap mapper(parameters_);
  mapper.setColorModels(objectModel_,nonObjectModel_);

  // now, just generate the maps
  // can be
  // - an image
  // - a camera stream
  source src(sourceType_,imgFiles);

  lti::image img; 
  lti::channel cmap;
  lti::viewer2D viewOrg("Input"),viewMap("Probability");
  lti::viewer2D::interaction action;
  lti::ipoint position;
  bool canSave=true;

  // as long as there are available images
  while (src.next(img)) {
    mapper.apply(img,cmap);
    viewOrg.show(img);
    viewMap.show(cmap);
    if (pause_) {
      viewMap.waitInteraction(action,position);
    } else {
      viewMap.getLastAction(action,position);
    }
    if (action.action == lti::viewer2D::Closed) {
      break;
    } else if (action.action == lti::viewer2D::KeyPressed) {
      if ( (action.key == 'S') || (action.key == 's') ) {
        if (canSave) {
          src.dump();
          canSave = false;
        }
      }
    } else if (action.action == lti::viewer2D::KeyReleased) {
      if ( (action.key == 'S') || (action.key == 's') ) {
        if (!canSave) {
          canSave = true;
        }
      }
    }
  }
  
  viewOrg.hide();
  viewMap.hide();

  lti::passiveWait(10000); // wait 0.01 s for the viewer to gracefully end

  return true;
}


int main(int argc,char *argv[]) {

  // report all possible errors always, since cameras may not work
  static  lti::statusMonitorCerr sm;
  lti::setStatusMonitor(sm);

  try {
    example obj;
    return obj(argc,argv);
  }
  catch (lti::exception& exp) {
    std::cout << "An LTI::EXCEPTION was thrown: ";
    std::cout << exp.what() << std::endl;
  }
  catch (std::exception& exp) {
    std::cout << "std::exception was thrown: ";
    std::cout << exp.what() << std::endl;
  }
  catch (...) {
    std::cout << "Unknown exception thrown!" << std::endl;
  }
  return EXIT_FAILURE;

} // main
