/*
 * Copyright (C) 2007 by Pablo Alvarado
 * 
 * This file is part of the LTI-Computer Vision Library 2 (LTI-Lib-2)
 *
 * The LTI-Lib-2 is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the authors nor the names of its contributors may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/** 
 * \file   fastHessianDetection.cpp
 *         Contains an example of use for the class lti::fastHessianDetection.
 * \author Pablo Alvarado
 * \date   04.11.2007
 * revisions ..: $Id: fastHessianDetection.cpp,v 1.2 2007/11/11 19:42:42 alvarado Exp $
 */



// LTI-Lib Headers
#include "ltiObject.h"
#include "ltiMath.h"     // General lti:: math and <cmath> functionality
#include "ltiConfig.h"   // To check if GTK is there
#include "ltiTimer.h"    // To measure time
#include "ltiIOImage.h"  // To read images
#include "ltiChannel8.h" // Monochromatic byte channels
#include "ltiChannel32.h"// Integer channels
#include "ltiFastHessianDetection.h" // Integral image
#include "ltiMatrixTransform.h" // For rotation and scaling
#include "ltiSplitImageTorgI.h"
#include "ltiPassiveWait.h"
#include "ltiUniformDiscreteDistribution.h"

#include "ltiFastHessianDetection.h"
#include "ltiLispStreamHandler.h"
#include "ltiDraw.h"


#if HAVE_GTK
#include "ltiViewer2D.h" // The normal viewer
typedef lti::viewer2D viewer_type;
#else
#include "ltiExternViewer2D.h"  // An emergency exit to display images.
typedef lti::externViewer2D viewer_type;
#endif

// Standard Headers
#include <cstdlib>
#include <iostream>
#include <string>
#include <fstream>

using std::cout;
using std::cerr;
using std::endl;

enum {
  POSITION = 1,
  ROTATION = 2,
  SCALE    = 4
};

/*
 * Help 
 */
void usage() {
  cout << "Usage: fastHessianDetection image [options] [-h]" << endl;
  cout << "Find the locations for the given image, if one is given.\n";
  cout << "An empty image name means to start a test for the location \n";
  cout << "finder, for position, scale and angle of structures.\n" << endl;
  cout << "The file \"fastHessianDetection.dat\" allows to configure the\n";
  cout << "evaluated behaviour.\n" << endl;
  cout << "  -h show this help." << endl;
  cout << "  -r check rotation." << endl;
  cout << "  -p check position." << endl;
  cout << "  -s check scale." << endl;
}

/*
 * Parse the line command arguments
 */
void parseArgs(int argc, char*argv[], 
               std::string& filename,
               lti::ubyte& modes) {
  
  modes = 0;

  filename.clear();
  // check each argument of the command line
  for (int i=1; i<argc; i++) {
    if (*argv[i] == '-') {
      switch (argv[i][1]) {
        case 'h':
          usage();
          exit(EXIT_SUCCESS);
          break;
        case 'r':
          modes |= ROTATION;
          break;
        case 'p':
          modes |= POSITION;
          break;
        case 's':
          modes |= SCALE;
          break;
        default:
          break;
      }
    } else {
      filename = argv[i]; // guess that this is the filename
    }
  }

  if (modes == 0) {
    modes = (ROTATION | POSITION | SCALE);
  }

}

void checkPosition(lti::fastHessianDetection& detector) {
  // With a 11 pixel wide point, move it and extract the locations

  lti::channel8 chnl(255,255,lti::ubyte(0)); // black background
  lti::image canvas;
  lti::uniformDiscreteDistribution rnd(0,255);
  lti::draw<lti::ubyte> painter;
  painter.setColor(255);
  painter.use(chnl);

  static viewer_type viewer("Position check");

  lti::draw<lti::rgbaPixel> cpainter;
  cpainter.use(canvas);
  cpainter.setColor(lti::rgbaPixel(240,128,128));

  lti::list<lti::location> locs;

  // try 25 points
  int i;
  for (i=0;i<25;++i) {
    lti::ipoint pos(rnd.rand(),rnd.rand());
    painter.circle(pos,5,true); // filled circle
    
    detector.apply(chnl,locs);
    
    canvas.castFrom(chnl);
    cpainter.set(locs,true);

    viewer.show(canvas);
    
    chnl.fill(lti::ubyte(0));
  }
  
}

void checkScale(lti::fastHessianDetection& detector) {

  // Paint a growing point and check what ratio is produced.

  lti::channel8 chnl(255,255,lti::ubyte(0)); // black background
  lti::image canvas;
  lti::draw<lti::ubyte> painter;
  painter.setColor(255);
  painter.use(chnl);

  static viewer_type viewer("Scale check");

  lti::draw<lti::rgbaPixel> cpainter;
  cpainter.use(canvas);
  cpainter.setColor(lti::rgbaPixel(240,128,128));

  lti::list<lti::location> locs;

  // try some scalings
  int i;
  for (i=2;i<50;i+=1) {
    painter.circle(127,127,i,true); // filled circle
    
    detector.apply(chnl,locs);
    
    canvas.castFrom(chnl);
    cpainter.set(locs,true);

    viewer.show(canvas);
    
    chnl.fill(lti::ubyte(0));
  }
  
}

void checkRotation(lti::fastHessianDetection& detector) {

  // With a 11 pixel wide point, move it and extract the locations
  static const int radius = 21;
  static const int width  = 5;


  lti::channel8 chnl(255,255,lti::ubyte(0)); // black background
  lti::image canvas;
  lti::draw<lti::ubyte> painter;
  painter.setColor(255);
  painter.use(chnl);

  static viewer_type viewer("Rotation check");

  lti::draw<lti::rgbaPixel> cpainter;
  cpainter.use(canvas);
  cpainter.setColor(lti::rgbaPixel(240,128,128));

  lti::list<lti::location> locs;

  // try some scalings
  int i;
  for (i=0;i<=360;i+=10) {

    painter.setColor(128);
    painter.ellipse(lti::ipoint(127,127),radius,width,
                    float(lti::degToRad((i+90)%360)),true); 

    painter.setColor(255);
    painter.circle(127,127,lti::iround(width*1.4),true); // filled circle

//     painter.rectangle(127-width,127-radius,127+width,127+radius,
//                       float(lti::degToRad((i+90)%360)),false);
//     painter.fill(127,127);

    detector.apply(chnl,locs);
    
    canvas.castFrom(chnl);
    cpainter.set(locs,true);

    viewer.show(canvas);
    
    chnl.fill(lti::ubyte(0));
  }
  
}


/*
 * Check all properties of the detector
 */
bool check(lti::fastHessianDetection& detector,lti::ubyte& modes) {
  if (modes & POSITION) {
    checkPosition(detector);
  }
  if (modes & SCALE) {
    checkScale(detector);
  }
  if (modes & ROTATION) {
    checkRotation(detector);
  }
  return true;
}

/*
 * Main method
 */
int main(int argc, char* argv[]) {

  std::string file;
  lti::ubyte modes;
  parseArgs(argc,argv,file,modes);

  static const char* confFile = "fastHessianDetection.dat";

  lti::fastHessianDetection::parameters param;
 
  // try to read the configuration file
  std::ifstream in(confFile);
  bool write=true;
  if (in) {
    lti::lispStreamHandler lsh;
    lsh.use(in);
    write=!param.read(lsh);
  }
  if (write) {
    // something went wrong reading, write a new configuration file
    std::ofstream out(confFile);
    lti::lispStreamHandler lsh;
    lsh.use(out);
    param.write(lsh);
  }
  
  // create the detector with the user specified configuration
  lti::fastHessianDetection detector(param);

  if (file.empty()) {
    if (check(detector,modes)) {
      std::cout << "Press Enter to continue" << std::endl;
      getchar();
      return EXIT_SUCCESS;
    } else {
      return EXIT_FAILURE;
    }
  }

  lti::image img;
  lti::channel8 chnl;
  lti::channel det;
  lti::ioImage loader;
  lti::splitImageTorgI splitter;
  
  if (!loader.load(file,img)) {
    std::cout << loader.getStatusString() << std::endl;
    return EXIT_FAILURE;
  }

  splitter.getIntensity(img,chnl);
  lti::list<lti::location> locs;
  lti::list<float> strengths;
  detector.apply(chnl,locs,strengths);
    
  // paint the locations
  lti::image canvas;
  canvas.castFrom(chnl);
  lti::draw<lti::rgbaPixel> painter;
  painter.use(canvas);
  lti::list<lti::location>::const_iterator it;
  
  for (it=locs.begin();it!=locs.end();++it) {
    painter.setColor(lti::rgbaPixel(255,192,100));
    lti::location loc(*it);
    loc.radius*=1.5; // param.orientationNeighborhoodFactor;
    painter.set(loc,true);
  }

  viewer_type view("Original");
  viewer_type viewd("Locations");

  view.show(chnl);
  viewd.show(canvas);
    
  std::cout << "Press Enter to continue" << std::endl;
  getchar();

  return EXIT_SUCCESS;
}
