/*
 * Copyright (C) 2007-2008
 * Pablo Alvarado
 *
 * This file is part of the LTI-Computer Vision Library 2 (LTI-Lib-2)
 *
 * The LTI-Lib-2 is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the authors nor the names of its contributors may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * \file   fastHessianDetectionOptimization.cpp
 *         Utility to find the pareto front of the fastHessianDetection.
 * \author Pablo Alvarado
 * \date   18.12.07
 *
 * $Id: fastHessianDetectionOptimization.cpp,v 1.1 2007/12/26 16:32:09 alvarado Exp $
 */


/*
 * Pareto Analysis Tool for lti::fastHessianDetection
 */

#include "ltiFastHessianDetectionGenetics.h"
#include "ltiParetoFront.h"
#include "ltiStreamProgressInfo.h"

#include "ltiDraw.h"
#include "ltiViewer2D.h"
#include "ltiImage.h"
#include "ltiColors.h"

#include <iostream>
#include <fstream>

using std::cout;
using std::cerr;
using std::endl;

namespace lti {

  /**
   * The real class
   */
  class tester {
  public:
    void operator()(int argc,char *argv[]);
  };

  /**
   * Pareto Front domination
   */
  bool dominate(const dvector& a,
                const dvector& b,
                const ivector& idx) {
    bool theOne = false;
    
    // a little pointer arithmetic to accelerate this frequently called
    // function
    int i,j;
    for (i=0;i<idx.size();++i) {
      j=idx.at(i);
      if (a[j] < b[j]) {
        // if any element is smaller => definitively not greater!
        return false;
      } else if (a[j] > b[j]) {
        // only greater if at least one element has been strictly greater
        theOne = true;
      }
    }
    
    return theOne;
  }
  
  /**
   * Get non-dominated elements in the matrix
   * Return the size of the front (num of elements with true
   */
  int getNonDominated(const dmatrix& mat,
                      const ivector& idx,
                      std::vector<bool>& nonDom) {
    int i,j;
    bool dominated;
    int n = 0;
    nonDom.resize(mat.rows());
    for (i=0;i<mat.rows();++i) {
      dominated = false;
      for (j=0;!dominated && (j<mat.rows());++j) {
        dominated = dominate(mat.getRow(j),mat.getRow(i),idx);
      }
      nonDom[i] = !dominated;
      if (nonDom[i]) {
        n++;
      }
    }

    return n;
  }
  
  void plotIndividuals(const dmatrix& bbox,
                       const dmatrix& indivs,
                       image& canvas,
                       const rgbaPixel& color = White,
                       const char symbol = 'o',
                       const bool line=false) {
    
    static const int border = 8;

    draw<rgbaPixel> painter;
    painter.use(canvas);

    // set colors and symbols
    painter.setColor(color);
    std::string str = "o";
    str[0]=symbol;
    painter.setStyle(str.c_str());

    double mx,my,bx,by;
    int j;
    mx = (canvas.lastColumn()-2*border)/(bbox.at(1,0)-bbox.at(0,0));
    bx = border-bbox.at(0,0)*mx;

    my = (canvas.lastRow()-2*border)/(bbox.at(0,1)-bbox.at(1,1));
    by = border-bbox.at(1,1)*my;
    
    ipoint p;
    const rgbaPixel lineColor(64,128,255);

    if ((indivs.rows() > 0) && line) {
      p.x = iround(mx*indivs.at(0,0) + bx);
      p.y = iround(my*indivs.at(0,1) + by);
      
      painter.setColor(lineColor);
      painter.set(p);
      painter.setColor(color);
    }


    for (j=0;j<indivs.rows();++j) {
      p.x = iround(mx*indivs.at(j,0) + bx);
      p.y = iround(my*indivs.at(j,1) + by);

      if (line) {
        painter.setColor(lineColor);
        painter.lineTo(p);
        painter.setColor(color);        
      }

      painter.marker(p);
    }
  };

  // read pareto front file 
  bool 
  readParetoFront(const std::string& file,
                  std::vector<dvector>& subFitness,
                  std::vector<fastHessianDetection::parameters>& parSet) {
    std::ifstream in(file.c_str());
    if (in) {
      lispStreamHandler lsh(in);

      parSet.clear();
      parSet.reserve(250);

      subFitness.clear();
      subFitness.reserve(250);

      // read all parameterization into parSet
      while (lsh.readBegin()) {
        // two possibilities:
        // 1. BoundingBox symbol (the last one), or
        // 2. A list with the vector of interest and the parameterization
        if (lsh.tryBegin()) {
          // the vector
          subFitness.push_back(dvector());
          subFitness.back().read(lsh,false); // the open parenthesis was
          // already read
          lsh.readEnd();
          
          // the parameters we care on
          parSet.push_back(fastHessianDetection::parameters());
          if (!parSet.back().read(lsh)) {
            std::cerr << "Error reading parameterization: \n" 
                      << lsh.getStatusString() << std::endl;
          }
          
          lsh.readEnd(); // of the vector-params block
        } else {
          // read the bounding box and end
          dmatrix bbox;
          lti::read(lsh,"BoundingBox",bbox,false);
          lsh.readEnd();
          break;
        }
      }
    } else {
      return false;
    }

    in.close();
   
    if (parSet.empty()) {
      std::cerr << "No parameterization found in " << file << std::endl;
      return false;
    }
    
    return true;
  }
  
  // create file with parameters table for the given parameters
  bool createParamTable(const std::string& paretoFile) {
    std::vector<fastHessianDetection::parameters> parSet;
    std::vector<dvector> subFitness;
    
    std::cout << "Creating Parameters Table..."; 
    std::cout.flush();

    // try to open the paretoFile
    if (!readParetoFront(paretoFile.c_str(),subFitness,parSet)) {
      return false;
    }
    
    std::string tableFile = paretoFile + ".table.csv";
    std::ofstream o(tableFile.c_str());

    if (!o) {
      std::cerr << "File " << tableFile << " could not be created.\n";
      return false;
    }

    unsigned int i,j;
    // simple header, to know what is what!
    o << "f1";
    for (i=1;i<static_cast<unsigned int>(subFitness.front().size());++i) {
      o << "\tf" << i+1;
    }

    o << "\t#Levels";
    o << "\tiniKSize";
    o << "\tselMethod";
    o << "\tiniKStep";
    o << "\tgroupSize";
    o << "\tnormPower";
    o << "\textrema";
    o << "\tselMode";
    o << "\tthresh";
    o << "\tsubsample";
    o << "\tiniSampl";
    o << "\torientat";
    o << "\toneighb";
    o << "\tosampst";
    o << "\towavesf";
    o << "\tgaussf";
    o << "\towndwd";
    o << "\to#wnd";
    o << std::endl;

    for (i=0;i<parSet.size();++i) {
      // fitness

      o << subFitness[i].at(0);
      for (j=1;j<static_cast<unsigned int>(subFitness[i].size());++j) {
        o << "\t" << subFitness[i].at(j);
      }

      const fastHessianDetection::parameters& p = parSet[i]; 

      o << "\t" << p.numberOfLevels;
      o << "\t" << p.initialKernelSize;
      if (p.levelSelectionMethod == fastHessianDetection::Blocks) {
        o << "\t" << "Blocks";
      } else {
        o << "\t" << "Exponential";
      }
      o << "\t" << p.initialKernelStep;
      o << "\t" << p.levelGroupSize;
      o << "\t" << p.normPower;
      o << "\t";
      switch(p.extrema) {
        case fastHessianDetection::Minima:
          o << "Minima";
          break;
        case fastHessianDetection::Maxima:
          o << "Maxima";
          break;
        case fastHessianDetection::Both:
          o << "Both";
          break;
        default:
          o << "Unknown";
      } 
      o << "\t";
      switch(p.locationSelectionMode) {
        case fastHessianDetection::All:
          o << "All";
          break;
        case fastHessianDetection::Absolute:
          o << "Absolute";
          break;
        case fastHessianDetection::Relative:
          o << "Relative";
          break;
        case fastHessianDetection::Conspicuous:
          o << "Conspicuous";
          break;
        case fastHessianDetection::Number:
          o << "Number";
          break;
        default:
          o << "Unknown";
      }
      o << "\t" << p.threshold;
      o << "\t" << p.subsampleLevels;
      o << "\t" << p.initialSamplingStep;
      o << "\t" << p.computeOrientation;
      o << "\t" << p.orientationNeighborhoodFactor;
      o << "\t" << p.orientationSamplingStepFactor;
      o << "\t" << p.orientationWaveletSizeFactor;      
      o << "\t" << p.orientationGaussianFactor;
      o << "\t" << p.orientationWindowWidth;
      o << "\t" << p.orientationNumberOfWindows;
      
      o << std::endl;
    }

    o.close();

    std::cout << "done." << std::endl;
    return true;
  }


  // -------------------------------------------------------------------------
  // tester::operator() : The main program
  // -------------------------------------------------------------------------

  void tester::operator()(int argc,char *argv[]) {

    bool showProgress = true;
    bool showViewer = false;
    bool force = false;
    bool resume = false;
    bool createConfig = true;
    int verboseLevel = 1;
    std::string str;
    int i=0;
    std::string configFile = "config_fastHessianDetection.dat";
    std::string paretoFile = "";
    std::string paretoTable = "";

    // simple command line parsing
    while (i<argc) {
      str = argv[i];
      if ((str == "-h") || (str == "--help")) {
        std::cout << "Evaluation of fast hessian detection algorithm" 
                  << std::endl;
        std::cout << "Usage: " << argv[0] 
                  << " [-f] [-h] [-q] [-v] [-c configFile]\n\n";
        std::cout << "  -h This help.\n";
        std::cout << "  -f Force creation, even if files already exist.\n";
        std::cout << "  -v At the end of one evaluation show the front.\n";
        std::cout << "  -r Reassume a broken process.\n";
        std::cout << "  -l [012] Verbose level.\n";
        std::cout << "  -c config_file Use given configuration file.\n";
        std::cout << "  -e paretoFrontFile evaluate parameterizations.\n";
        std::cout << "  -t paretoFrontFile generate a parameter table.\n";
        std::cout << "  -q Do not output the progress of the evaluation.\n";
        return;
      } else if (str == "-q") {
        showProgress = false;
      } else if (str == "-v") {
        showViewer = true;
      } else if (str == "-f") {
        force = true;
      } else if (str == "-r") {
        resume = true;
      } else if (str == "-l") {
        ++i;
        str = argv[i];
        if (!str.empty()) {
          verboseLevel = str[0]-'0';
          std::cout << "Verbose level: " << verboseLevel << std::endl;
        }
      } else if (str.substr(0,2) == "-l") {
        verboseLevel = str[2]-'0';
        std::cout << "Verbose level: " << verboseLevel << std::endl;        
      } else if (str == "-c") {
        ++i;
        if ((i<argc) && (argv[i][0]!='-')) {
          configFile = argv[i];
          createConfig = false;
        }
      } else if (str == "-e") {
        ++i;
        if ((i<argc) && (argv[i][0]!='-')) {
          paretoFile = argv[i];
        }        
      } else if (str == "-t") {
        ++i;
        if ((i<argc) && (argv[i][0]!='-')) {
          paretoTable = argv[i];
        }        
      } 
      
      ++i;
    }

    if (!paretoTable.empty()) {
      createParamTable(paretoTable);
      return;
    }
    paretoFront::parameters pfPar;
    fastHessianDetectionGenetics::parameters fhdgPar;

    std::ifstream in(configFile.c_str());
    bool write=true;
    if (in) {
      lispStreamHandler lsh;
      lsh.use(in);
      write=!(lti::read(lsh,"showProgress",showProgress) &&
              lti::read(lsh,"verboseLevel",verboseLevel) &&
              lti::read(lsh,"pareto",pfPar) &&
              lti::read(lsh,"genetics",fhdgPar));
    }
    if (write) {
      std::ofstream out(configFile.c_str());
      lispStreamHandler lsh;
      lsh.use(out);
      lti::write(lsh,"showProgress",showProgress);
      lti::write(lsh,"verboseLevel",verboseLevel);
      lti::write(lsh,"pareto",pfPar);
      lti::write(lsh,"genetics",fhdgPar);

      std::cout << "Please edit file " << configFile 
                << " and ensure that the list of images is available."
                << std::endl;
      return;
    }

    // first a genetics object, with its corresponding parameters
    fastHessianDetectionGenetics fhdg(fhdgPar);
    // tell the parameters of the pareto front to use the given genetics object
    pfPar.setGeneticsObject(fhdg);
    // create a pareto object with the proper parameters
    paretoFront pf(pfPar);

    // we need to visualize progress
    streamProgressInfo progInfo;
    pf.setProgressObject(progInfo);
    pf.getProgressObject().setDetailLevel(verboseLevel);

    // and now, find the pareto front
    dmatrix front;
    pf.apply(front);

  }
}

int main(int argc,char *argv[]) {
  try {
    lti::tester testObj;
    testObj(argc,argv);
    std::cout << "Test at the end!" << std::endl;
  }
  catch (lti::exception& exp) {
    std::cout << "An LTI::EXCEPTION was thrown: ";
    std::cout << exp.what() << std::endl;
  }
  catch (std::exception& exp) {
    std::cout << "std::exception was thrown: ";
    std::cout << exp.what() << std::endl;
  }
  catch (...) {
    std::cout << "Unknown exception thrown!" << std::endl;
  }
  return 0;

} // main
