/*
 * Copyright (C) 2007 by Pablo Alvarado
 * 
 * This file is part of the LTI-Computer Vision Library 2 (LTI-Lib-2)
 *
 * The LTI-Lib-2 is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the authors nor the names of its contributors may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/** 
 * \file   surfLocalDescriptor.cpp
 *         Contains an example of use for the class lti::surfLocalDescriptor
 * \author Pablo Alvarado
 * \date   04.11.2007
 * revisions ..: $Id: fft.cpp,v 1.3 2012/06/20 03:04:34 alvarado Exp $
 */

// Standard Headers: from ANSI C and GNU C Library
#include <cstdlib>  // Standard Library for C++
#include <getopt.h> // Functions to parse the command line arguments

// Standard Headers: STL
#include <iostream>
#include <string>
#include <fstream>

// LTI-Lib Headers
#include "ltiIOImage.h"
#include "ltiMath.h"

#include "ltiLispStreamHandler.h"
#include "ltiFFT.h"
#include "ltiIFFT.h"

#include "ltiViewer2D.h" // The normal viewer
#include "ltiRound.h"

// Standard Headers
#include <cstdlib>
#include <iostream>
#include <string>
#include <fstream>

using std::cout;
using std::cerr;
using std::endl;


/*
 * Help 
 */
void usage()  {
  
  cout <<
    "usage: fft [<file>]\n\n"   \
    "       -h|--help        show this help\n"\
    "       <file>           optional input image" << std::endl;    
}

/*
 * Parse the line command arguments
 * 
 * Set the filename with the given parameter, or an empty string otherwise
 */
void parse(int argc, char*argv[],
           std::string& filename) {

  // We use the standard getopt.h functions here to parse the arguments.
  // Check the documentation of getopt.h for more information on this
  
  // structure for the long options. 
  static struct option lopts[] = {
    {"help",no_argument,0,'h'},
    {0,0,0,0}
  };

  int optionIdx;
  int cc;

  while ((cc = getopt_long(argc, argv, "h", lopts,&optionIdx)) != -1) {
    switch (cc) {
    case 'h':
      usage();
      exit(EXIT_SUCCESS); 
    default:
      cerr << "Option '-" << static_cast<char>(cc) << "' not recognized." 
           << endl;
    }
  }

  // Now try to read the image name
  if (optind < argc) {           // if there are still some arguments left...
    filename = argv[optind];  // with the given image file
  } else {
    filename = "";
  }
}

/*
 * Main method
 */
int main(int argc, char* argv[]) {

  // check if the use gave some filename 
  std::string filename;
  parse(argc,argv,filename);

  const int size = 256;
  lti::channel frec(size,size,0.0f),space(size,size,0.0f),phase(size,size,0.0f);

  if (filename.empty()) {
    lti::viewer2D viewF("Frecuency");
    lti::viewer2D viewS("Space");
    
    lti::viewer2D::parameters vpar;
    vpar.mappingType = lti::viewer2DPainter::Optimal;
    vpar.title = "Space";
    viewS.setParameters(vpar);
    
    lti::viewer2D::interaction action;
    lti::ipoint pos;
    
    lti::ifft::parameters ifftPar;
    ifftPar.mode = lti::Polar;
    lti::ifft iFFT(ifftPar);
    
    std::cout << "Just press the button on the Frequency window." << std::endl;
    
    do {
      viewF.show(frec);
      viewF.waitButtonPressed(action,pos);
      float zoom = viewF.getParameters().zoom;
      
      lti::fpoint fpos(pos);
      fpos/=zoom;
      
      frec.fill(0.0f);
      frec.at(static_cast<int>(fpos.y),static_cast<int>(fpos.x))=1.0f;
      
      iFFT.apply(frec,phase,space);
      
      viewS.show(space);
      
    } while(action.action != lti::viewer2D::Closed);
  } else {
        // we need an object to load images
    lti::ioImage loader;

    // we also need an image object
    lti::image img;

    // gray scale image
    lti::channel chnl;

    // load the image
    if (loader.load(filename,img)) {
      // image successfully loaded
      chnl.castFrom(img);

      // we need a viewer 
      static lti::viewer2D view(filename);
      static lti::viewer2D viewmag("Magnitude");
      static lti::viewer2D viewarg("Argument");
      
      view.show(chnl);

      lti::fft::parameters fftPar;
      fftPar.mode = lti::Polar;
      lti::fft FFT(fftPar);
      
      FFT.apply(chnl,frec,phase);
      viewmag.show(frec);
      viewarg.show(phase);

      lti::viewer2D::interaction action;
      lti::ipoint pos;
      do {
        view.waitInteraction(action,pos);
      } while(action.action != lti::viewer2D::Closed);
    }
  }

  return EXIT_SUCCESS;
}
