/*
 * Copyright (C) 2009-2012 by Pablo Alvarado
 * 
 * This file is part of the course MP-6123 Digital Image Processing,
 * at the Costa Rica Institute of Technology.
 *
 * The LTI-Lib-2 is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the authors nor the names of its contributors may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/** 
 * \file   Frequency domain examples
 *         Beamer class image generation tool
 * \author Pablo Alvarado
 * \date   26.05.2012
 */

// Standard Headers: from ANSI C and GNU C Library
#include <cstdlib>  // Standard Library for C++
#include <getopt.h> // Functions to parse the command line arguments

// Standard Headers: STL
#include <iostream>
#include <string>
#include <fstream>

// LTI-Lib Headers
#include <ltiObject.h>
#include <ltiMath.h>     // General lti:: math and <cmath> functionality
#include <ltiTimer.h>    // To measure time

#include <ltiIOImage.h>  // To read/write images from files (jpg, png, and bmp)
#include <ltiViewer1D.h>
#include <ltiViewer2D.h>

#include <ltiLispStreamHandler.h>
#include <ltiGaussKernels.h>
#include <ltiFFT.h>
#include <ltiIFFT.h>
#include <ltiRound.h>
#include <ltiBoundaryExpansion.h>

// Ensure that the STL streaming is used.
using std::cout;
using std::cerr;
using std::endl;


/*
 * Help 
 */
void usage()  {

  
  cout <<
    "usage: code [options] [<file>]\n\n"   \
    "       -h|--help        show this help\n"\
    "       <file>           input image" << std::endl;    
  cout << 
    "The file 'freq.dat' will be read for further options including " \
    "the filter type,\n radius, order, etc.\n\n"\
    "The radius must be given in normalized frequency, i.e. it must be "\
    "positive and less than 0.5\n\n" \
    "Press '?' over the 'Freq response' viewer for some keyboard interaction\n"\
       << std::endl;
}

enum eUpdateMode {
  Radius,
  Order
};

enum eFilterType {
  None,
  LPGauss,
  LPIdeal,
  LPButterworth,
  HPGauss,
  HPIdeal,
  HPButterworth,
  Laplacian,
  DoG
};

std::string toStr(const eFilterType t) {
  switch(t) {
  case None:
    return "None";
    break;
  case LPGauss:
    return "LPGauss";
    break;
  case LPIdeal:
    return "LPIdeal";
    break;
  case LPButterworth:
    return "LPButterworth";
    break;
  case HPGauss:
    return "HPGauss";
    break;
  case HPIdeal:
    return "HPIdeal";
    break;
  case HPButterworth:
    return "HPButterworth";
    break;
  case Laplacian:
    return "Laplacian";
    break;
  case DoG:
    return "DoG";
    break;
  default:
    return "Unknown";
  }
}

namespace lti {

  bool read(ioHandler& handler,
            eFilterType& type) {
    bool b=true;

    std::string str;
    b = b && lti::read(handler,str);
    if (str.find("None") != std::string::npos) {
      type = None;
    } else if (str.find("LPGauss") != std::string::npos) {
      type = LPGauss;
    } else if (str.find("LPIdeal") != std::string::npos) {
      type = LPIdeal;
    } else if (str.find("LPButterworth") != std::string::npos) {
      type = LPButterworth;
    } else if (str.find("HPGauss") != std::string::npos) {
      type = HPGauss;
    } else if (str.find("HPIdeal") != std::string::npos) {
      type = HPIdeal;
    } else if (str.find("HPButterworth") != std::string::npos) {
      type = HPButterworth;
    } else if (str.find("DoG") != std::string::npos) {
      type = DoG;
    } else if (str.find("Laplacian") != std::string::npos) {
      type = Laplacian;
    }
  
    return b;
  }
  
  bool write(ioHandler& handler,
             const eFilterType& type) {
    return lti::write(handler,toStr(type));
  }
}

/*
 * Parse the line command arguments
 * 
 * Set the filename with the given parameter, or an empty string otherwise
 */
void parse(int argc, char*argv[],
           std::string& filename) {

  // We use the standard getopt.h functions here to parse the arguments.
  // Check the documentation of getopt.h for more information on this
  
  // structure for the long options. 
  static struct option lopts[] = {
    {"help",no_argument,0,'h'},
    {0,0,0,0}
  };

  int optionIdx;
  int cc;

  while ((cc = getopt_long(argc, argv, "h", lopts,&optionIdx)) != -1) {
    switch (cc) {
    case 'h':
      usage();
      exit(EXIT_SUCCESS); 
    default:
      cerr << "Option '-" << static_cast<char>(cc) << "' not recognized." 
           << endl;
    }
  }

  // Now try to read the image name
  if (optind < argc) {           // if there are still some arguments left...
    filename = argv[optind];  // with the given image file
  } else {
    filename = "";
  }
}


/*
 * Main method
 */
int main(int argc, char* argv[]) {
  
  // check if the use gave some filename 
  std::string filename;
  parse(argc,argv,filename);

  // parameters for our mapping object
  static const char* fname = "freq.dat";

  // Desired filter type
  eFilterType type = LPIdeal;
  int order = 2;
  float radius(.25f);

  // try to read them from disk
  bool write=true;
  std::ifstream in(fname);
  lti::eBoundaryType padding=lti::Zero;

  if (in) {
    lti::lispStreamHandler lsh(in);
    write= !(lti::read(lsh,"type",type) &&
             lti::read(lsh,"order",order) &&
             lti::read(lsh,"radius",radius) &&
             lti::read(lsh,"boundary",padding));
  } 
  if (write) {
    std::ofstream out(fname);
    lti::lispStreamHandler lsh(out);
    lti::write(lsh,"type",type);
    lti::write(lsh,"order",order);
    lti::write(lsh,"radius",radius);
    lti::write(lsh,"boundary",padding);
    out << std::endl;
    out.close();
  }


  if (!filename.empty()) {
    // we need an object to load images
    lti::ioImage loader;

    // we also need an image object
    lti::image img;

    // we need viewers
    static lti::viewer2D view(filename);
    static lti::viewer2D viewh("PSF",lti::viewer2DPainter::Optimal);
    static lti::viewer2D viewH("Freq response",
                               lti::viewer2DPainter::OptimalLog);
    static lti::viewer2D viewF("Image Spectrum",
                               lti::viewer2DPainter::OptimalLog);
    static lti::viewer2D viewHF("Filtered Spectrum",
                                lti::viewer2DPainter::OptimalLog);
    static lti::viewer2D viewf("Filtered",lti::viewer2DPainter::Optimal);

    lti::viewer1D::parameters v1par;
    v1par.drawHorizontalAxis=true;
    v1par.drawMeanValue=false;
    v1par.drawGrid=true;
    v1par.pixelsPerElement=1;
    v1par.style=lti::viewer1DPainter::Impulses;
    v1par.title="Profile H(wx,0)";
    static lti::viewer1D viewpH(v1par);
    
    v1par.title="Profile h(x,0)";
    static lti::viewer1D viewph(v1par);
    
    
    // load the image
    if (loader.load(filename,img)) {
      // image successfully loaded
      
      // ******************* 
      // **    PADDING    ** 

      // we need to pad it to next two-power
      const int n = lti::iround(lti::pow(2.0f,
                                         ceil(lti::log(2*lti::max(img.rows(),
                                                               img.columns()))/
                                              lti::log(2.0f))));
      const int nh = n/2;

      lti::boundaryExpansion::parameters bepar;
      bepar.boundaryType = padding;

      bepar.topBorder = 0;
      bepar.leftBorder = 0;
      bepar.bottomBorder = n-img.rows();
      bepar.rightBorder = n-img.columns();

      lti::boundaryExpansion be(bepar);

      // ready to pad
      lti::channel chnl;
      chnl.castFrom(img);
      be.apply(chnl);

      // ** END OF PADDING ** 
      // ******************** 

      // show the padded image
      view.show(chnl); 

      // shift the spectrum (in the space domain)
      for (int y=0;y<n;++y) {
        float s=lti::even(y) ? 1.f : -1.f;
        for (int x=0;x<n;++x) {
          chnl.at(y,x)*=s;
          s*=-1.f;
        }
      }

      // all partial results
      // h      PSF (impulse response)
      // H      Magnitude response
      // Hang   Phase response
      // Hre    Real part of spectrum of the space shifted PSF
      // Him    Imaginary part of spectrum of the space shifted PSF
      // F      Magnitude of spectrum of input channel
      // Fang   Angle of spectrum of F
      // Hf_abs Product of H and F
      // Hf_arg Phase of product of H and F
      lti::channel h,F,Fang,Hf_abs,Hf_arg,f;
      lti::channel H(n,n),Hang(n,n,0.0f),Hre(n,n),Him(n,n,0.0f);

      // compute the (shifted) spectrum of the original channel (shifted)
      static const lti::eCoordinateSystem csys = lti::Polar;
      lti::fft fft2D;
      fft2D.apply(chnl,csys,F,Fang);

      // show magnitude of the spectrum
      viewF.show(F);

      bool theEnd = false;
      eUpdateMode mode = Radius;
      bool highpass = false;
      
      // *****************************************************************
      // loop to allow user to change the radius, order, filter type, etc.
      // *****************************************************************
      do {
        const int R = lti::iround(radius*n); // radius in pixels (denormalized)
        const float R2 = lti::sqr(float(R));

        std::cout << "\nRadius: " << radius
                  << " (equivalent to " << R << "px)" << std::endl;
        std::cout << "Order: " << order << std::endl;
        std::cout << "Type:  " << toStr(type) << std::endl;

        // what to do
        switch (type) {

          /**
           * Forwards and Backwards FFT
           */
        case None: {
          // Frecuency response does not modify image
          H.assign(n,n,1.0f);
          Hang.assign(n,n,1.0f);
          Hre.assign(n,n,1.0f);
        } break;
        
        case LPIdeal: {
          // Compute H and Hre
          for (int y=0;y<n;++y) {
            float s=lti::even(y) ? 1.f : -1.f;
            for (int x=0;x<n;++x) {
              const float d2 = (lti::sqr(y-nh)+lti::sqr(x-nh)); 
              H.at(y,x)=(d2<=R2)?1.0f:0.0f;
              Hre.at(y,x)=H.at(y,x)*s;
              s*=-1.0f;
            }
          }
        } break;
        
        case LPGauss: {
          // Compute H and Hre
          for (int y=0;y<n;++y) {
            float s=lti::even(y) ? 1.f : -1.f;
            for (int x=0;x<n;++x) {
              const float d2 = (lti::sqr(y-nh)+lti::sqr(x-nh)); 
              H.at(y,x)=exp(-0.5*d2/R2);
              Hre.at(y,x)=H.at(y,x)*s;
              s*=-1.0f;
            }
          }
        } break;
        
        case LPButterworth: {
          // Compute H and Hre
          for (int y=0;y<n;++y) {
            float s=lti::even(y) ? 1.f : -1.f;
            for (int x=0;x<n;++x) {
              const float d2 = (lti::sqr(y-nh)+lti::sqr(x-nh)); 
              H.at(y,x)=1.0f/(1.0f+lti::pow(d2/R2,order));
              Hre.at(y,x)=H.at(y,x)*s;
              s*=-1.0f;
            }
          }
        } break;

        case HPGauss: {
          // Compute H and Hre
          for (int y=0;y<n;++y) {
            float s=lti::even(y) ? 1.f : -1.f;
            for (int x=0;x<n;++x) {
              const float d2 = (lti::sqr(y-nh)+lti::sqr(x-nh)); 
              H.at(y,x)=1.0f-exp(-0.5*d2/R2);
              Hre.at(y,x)=H.at(y,x)*s;
              s*=-1.0f;
            }
          }

        } break;

        case HPIdeal: {
          // Compute H and Hre
          for (int y=0;y<n;++y) {
            float s=lti::even(y) ? 1.f : -1.f;
            for (int x=0;x<n;++x) {
              const float d2 = (lti::sqr(y-nh)+lti::sqr(x-nh)); 
              H.at(y,x)=(d2>R2)?1.0f:0.0f;
              Hre.at(y,x)=H.at(y,x)*s;
              s*=-1.0f;
            }
          }

        } break;

        case HPButterworth: {
          // Compute H and Hre
          for (int y=0;y<n;++y) {
            float s=lti::even(y) ? 1.f : -1.f;
            for (int x=0;x<n;++x) {
              const float d2 = (lti::sqr(y-nh)+lti::sqr(x-nh)); 
              H.at(y,x)=1.0f/(1.0f+lti::pow(R2/d2,order));
              Hre.at(y,x)=H.at(y,x)*s;
              s*=-1.0f;
            }
          }

        } break;
        
        case DoG:
          break;

        case Laplacian: {
          // the values are real negative:
          Hang.assign(n,n,lti::Pi);

          // Compute H and Hre
          for (int y=0;y<n;++y) {
            float s=lti::even(y) ? 1.f : -1.f;
            for (int x=0;x<n;++x) {
              const float d2 = (lti::sqr(y-nh)+lti::sqr(x-nh)); 
              H.at(y,x)=4*lti::Pi*d2;
              Hre.at(y,x)=-H.at(y,x)*s;
              s*=-1.0f;
            }
          }

        } break;

        default:
          std::cout << "Unknown type" << std::endl;
        }

        // apply filter in the frequency domain
        Hf_abs.emultiply(F,H);
        Hf_arg.copy(Fang); // Ignore Hang, because it has a shift

        lti::ifft ifft2D;
        // compute the impulse response (PSF)
        ifft2D.apply(Hre,Hang,lti::Cartesian,h);
        // back to time
        ifft2D.apply(Hf_abs,Hf_arg,csys,f);
        // shift back the result (in the space domain)
        for (int y=0;y<f.rows();++y) {
          float s=lti::even(y) ? 1.f : -1.f;
          for (int x=0;x<f.columns();++x) {
            f.at(y,x)*=s;
            h.at(y,x)*=s; // shift back
            s*=-1.f;
          }
        }
      
        // show the results
        viewH.show(H);
        viewHF.show(Hf_abs);
        viewf.show(f);
        viewh.show(h);
        viewph.show(h.getRow(nh));
        viewpH.show(H.getRow(nh));


        // wait for the user to close the mapped window 
        lti::viewer2D::interaction action;
        lti::ipoint pos;

        viewH.waitInteraction(action,pos); // wait for something to happen
        if (action == lti::viewer2D::Closed) { // window closed?
          theEnd = true; // we are ready here!
        } else if (action == lti::viewer2D::KeyPressed) {
          switch(action.key) {
          case lti::viewer2D::UpKey:
            if (mode == Radius) {
              radius*=2;
            } else {
              order*=2;
            }
            break;
          case lti::viewer2D::DownKey:
            if (mode == Radius) {
              radius/=2;
            } else {
              order/=2;
            }
            break;
          case lti::viewer2D::LeftKey:
            if (mode == Radius) {
              radius-=1.0f/nh;
            } else {
              order-=1;
            }
            break;
          case lti::viewer2D::RightKey:
            if (mode == Radius) {
              radius+=1.0f/nh;
            } else {
              order+=1;
            }
            break;
          case 'B':
          case 'b':
            type = highpass? HPButterworth : LPButterworth;
            break;
          case 'G':
          case 'g':
            type = highpass? HPGauss : LPGauss;
            break;
          case 'I':
          case 'i':
            type = highpass? HPIdeal : LPIdeal;
            break;
          case 'L':
          case 'l':
            type = Laplacian;
            break;
          case 'R':
          case 'r':
            mode = Radius;
            std::cout << "Arrows will change radius of the filter" << std::endl;
            break;
          case 'O':
          case 'o':
            mode = Order;
            std::cout << "Arrows will change order of the filter" << std::endl;
            break;
          case 'H':
          case 'h':
            highpass = !highpass;
            if (highpass) {
              switch (type) {
              case LPButterworth:
                type=HPButterworth;
                break;
              case LPGauss:
                type=HPGauss;
                break;
              case LPIdeal:
                type=HPIdeal;
                break;
              default:
                break;
              }
            } else {
              switch (type) {
              case HPButterworth:
                type=LPButterworth;
                break;
              case HPGauss:
                type=LPGauss;
                break;
              case HPIdeal:
                type=LPIdeal;
                break;
              default:
                break;
              }
            }
            break;
          default:
            std::cout << 
              "Use:\n" \
              "  Arrow keys to modify filter radius or order\n" \
              "  B Butterworth filter\n" \
              "  G Gaussian filter\n" \
              "  I Ideal filter\n" \
              "  H Toggle between high-bass and low-pass\n" \
              "  R Set arrows to change radius\n" \
              "  O Set arrows to change order\n" \
              "  L Laplacian\n" << std::endl;
            break;
          }
          if (radius<=0) {
            radius = 1.0f/nh;
          } else if (radius>0.5f) {
            radius=0.5;
          }
          if (order<1) {
            order=1;
          } else if (order>256) {
            order=256;
          }
        }
      } while(!theEnd);
      
      return EXIT_SUCCESS;
    }
  }
  return EXIT_FAILURE;
}
