/*
 * Copyright (C) 1998, 1999, 2000, 2001
 * Lehrstuhl fuer Technische Informatik, RWTH-Aachen, Germany
 * Copyright (C) 2009-2012 by Pablo Alvarado
 * 
 * This file is part of the LTI-Computer Vision Library (LTI-Lib) created as
 * part of the course MP-6123 Digital Image Processing, at the Costa Rica
 * Institute of Technology.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the authors nor the names of its contributors may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */ 


// Standard Headers: from ANSI C and GNU C Library
#include <iostream>
#include <fstream>
#include <string>
#include <cstdlib>
#include <getopt.h> // Functions to parse the command line arguments

#include "ltiLine.h"

#include "ltiViewer2D.h"
#include "ltiCannyEdges.h"
#include "ltiHoughLineTransform.h"

#include "ltiIOImage.h"
#include "ltiLispStreamHandler.h"

#include "ltiDraw.h"
#include "ltiColors.h"

/*
 * Help 
 */
void usage()  {

  
  std::cout <<
    "usage: hough [options] [<file>]\n\n"   \
    "       -h|--help        show this help\n"\
    "       <file>           input image" << std::endl;    
  std::cout << 
    "The file 'hough.dat' will be read for further options.\n"
       << std::endl;
}

/*
 * Parse the line command arguments
 * 
 * Set the filename with the given parameter, or an empty string otherwise
 */
void parse(int argc, char*argv[],
           std::string& filename) {

  // We use the standard getopt.h functions here to parse the arguments.
  // Check the documentation of getopt.h for more information on this
  
  // structure for the long options. 
  static struct option lopts[] = {
    {"help",no_argument,0,'h'},
    {0,0,0,0}
  };

  int optionIdx;
  int cc;

  while ((cc = getopt_long(argc, argv, "h", lopts,&optionIdx)) != -1) {
    switch (cc) {
    case 'h':
      usage();
      exit(EXIT_SUCCESS); 
    default:
      std::cerr << "Option '-" << static_cast<char>(cc) << "' not recognized." 
                << std::endl;
    }
  }

  // Now try to read the image name
  if (optind < argc) {           // if there are still some arguments left...
    filename = argv[optind];  // with the given image file
  } 
}

/*
 * Main method
 */
int main(int argc, char* argv[]) {
  std::string file = "../../img/board.png";
  parse(argc,argv,file);

  lti::ioImage loader;
  lti::image img;
    
  if (!loader.load(file,img)) {
    std::cout << "Could not read file '" << file << "'" << std::endl;
    exit(1);
  }
  
  const char* cfgFile = "hough.dat";

  lti::houghLineTransform::parameters htPar;
  lti::cannyEdges::parameters cannyPar;
 
  std::ifstream in(cfgFile);
  bool w=true;
  
  if (in) {
    lti::lispStreamHandler lsh;
    lsh.use(in);
    w = ! (lti::read(lsh,"houghLineTransform",htPar) &&
           lti::read(lsh,"canny",cannyPar));
    
  }
  if (w) {
    std::ofstream out(cfgFile);
    lti::lispStreamHandler lsh;
    lsh.use(out);
    lti::write(lsh,"houghLineTransform",htPar);
    lti::write(lsh,"canny",cannyPar);
  }
  
  lti::houghLineTransform ht(htPar);
  
  lti::channel8 edges;
  lti::channel angles,strength; // orientation map
  
  // we need the edges
  lti::cannyEdges canny(cannyPar);
  canny.apply(img,edges,angles,strength);
  
  lti::viewer2D viewOrg(file); 
  viewOrg.show(img);

  lti::viewer2D viewEdge("Edges");

  lti::channel32 transformAccu;
  lti::genericVector<lti::iline> segs;
  lti::channel fedges;
  
  if (htPar.accumulationMode != lti::houghLineTransform::Classic) {

    fedges.castFrom(edges);
    fedges.emultiply(strength);

    // Compute Hough accumulation space
    // (this is not really necessary, but nice to see)
    if (!ht.apply(fedges,angles,transformAccu)) {
      std::cout << "Error in apply method of the houghLineTransform: " 
                << ht.getStatusString() << std::endl;
      return EXIT_FAILURE;
    }

    // The Hough Transform (this is the real call that provides the line segs)
    if (!ht.apply(fedges,angles,segs)) {
      std::cout << "Error in apply method of the houghLineTransform: " 
                << ht.getStatusString() << std::endl;
      return EXIT_FAILURE;
    }

  } else {
    fedges.castFrom(edges);

    // Compute Hough accumulation space
    // (this is not really necessary, but nice to see)
    if (!ht.apply(edges,angles,transformAccu)) {
      std::cout << "Error in apply method of the houghLineTransform: " 
                << ht.getStatusString() << std::endl;
      return EXIT_FAILURE;
    }

    // The Hough Transform (this is the real call that provides the line segs)
    if (!ht.apply(edges,angles,segs)) {
      std::cout << "Error in apply method of the houghLineTransform: " 
                << ht.getStatusString() << std::endl;
      return EXIT_FAILURE;
    }

  }
  
  lti::viewer2D viewHA("Hough Acc",lti::viewer2DPainter::Optimal);
  viewHA.show(transformAccu);
  
  // Draw the found lines in the original image
  lti::draw<lti::rgbaPixel> painter;
  painter.use(img);
  painter.setColor(lti::Green);
  for (int k=0;k<segs.size();++k) {
    painter.line(segs.at(k));
  }
    
  // at last, show the resulting lines
  lti::viewer2D view("Detected lines");
  view.show(img);
  
  lti::ipoint pos;
  lti::viewer2D::interaction action;
  bool doExit = false;

  lti::image overlay;
  overlay.castFrom(fedges);

  painter.use(overlay);
  painter.setColor(lti::rgbaPixel(255,128,0));
  viewEdge.show(fedges);

  do {
    viewHA.waitInteraction(action,pos); // wait for a key being pressed

    if (action.action == lti::viewerBase::ButtonPressed) {
      if (action.key == lti::viewerBase::LeftButton) {
        overlay.castFrom(fedges);
      }

      lti::iline seg;
      ht.accToSegment(transformAccu.size(),pos,overlay.size(),seg);
      painter.line(seg);
      
      viewEdge.show(overlay);
    } else if (action.action == lti::viewerBase::Closed) {
      doExit = true;
    }

  } while (!doExit);

  
  return EXIT_SUCCESS;
}
