/*
 * Copyright (C) 2007 by Pablo Alvarado
 * 
 * This file is part of the LTI-Computer Vision Library 2 (LTI-Lib-2)
 *
 * The LTI-Lib-2 is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the authors nor the names of its contributors may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

// LTI-Lib Headers
#include "ltiMath.h"     // General lti:: math and <cmath> functionality
#include "ltiConfig.h"   // To check if GTK is there
#include "ltiTimer.h"    // To measure time
#include "ltiIOImage.h"  // To read images
#include "ltiChannel8.h" // Monochromatic byte channels
#include "ltiChannel32.h"// Integer channels
#include "ltiIntegralImage.h" // Integral image
#include "ltiBoundaryExpansion.h" // For validation of the boundary conditions
#include "ltiUniformDiscreteDistribution.h" // Random number generation


#if HAVE_GTK
#include "ltiViewer2D.h" // The normal viewer
typedef lti::viewer2D viewer_type;
#else
#include "ltiExternViewer2D.h"  // An emergency exit to display images.
typedef lti::externViewer2D viewer_type;
#endif

// Standard Headers
#include <cstdlib>
#include <iostream>
#include <string>

using std::cout;
using std::cerr;
using std::endl;

/*
 * Help 
 */
void usage() {
  cout << "Usage: integralImage image [-s size] [-h]" << endl;
  cout << "Computes an average filter around each pixel with a window of " 
       << "a given size." << endl << endl;
  cout << "  image:   name of an image file, that has to be jpeg, png or bmp."
       << endl;
  cout << "           An empty image name means to start a test for the " 
       << endl;
  cout << "           selected boundary type." << endl;
  cout << "  -s size: the size of the window side being used for the average."
       << endl
       << "           [Default size is 11]" << endl;
  cout << "  -z: force zero boundary" << endl;
  cout << "  -c: force constant boundary" << endl;
  cout << "  -m: force mirror boundary"   << endl;
  cout << "  -p: force periodic boundary" << endl;
  cout << "  -h show this help." << endl;
}

/*
 * Parse the line command arguments
 */
void parseArgs(int argc, char*argv[], 
               std::string& filename,
               int& size,
               lti::eBoundaryType& btype) {
  filename.clear();
  size=11;
  btype = lti::NoBoundary;
  // check each argument of the command line
  for (int i=1; i<argc; i++) {
    if (*argv[i] == '-') {
      switch (argv[i][1]) {
        case 'h':
          usage();
          exit(EXIT_SUCCESS);
          break;
        case 's': // get the size from the next token
          size=lti::within(atoi(argv[++i]),1,65535);
          break;
        case 'z':
          btype = lti::Zero;
          break;
        case 'c':
          btype = lti::Constant;
          break;
        case 'm':
          btype = lti::Mirror;
          break;
        case 'p':
          btype = lti::Periodic;
          break;
        default:
          break;
      }
    } else {
      filename = argv[i]; // guess that this is the filename
    }
  }
}

/*
 * Check Constant boundary
 */

bool checkZero(const lti::matrix<lti::int32>& ii,
               const lti::integralImage& integrator) {

  bool res = true;
  // Input image was:
  //
  //  1  2  3  4
  //  5  6  7  8
  //  9 10 11 12
  // 13 14 15 16

  // the corners
  
  if (integrator.sum(ii,lti::ipoint(-2,-2),lti::ipoint(0,0)) != 1) {
    std::cerr << "Error in NW" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(3,-2),lti::ipoint(5,0)) != 4) {
    std::cerr << "Error in NE" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(-2,3),lti::ipoint(0,5)) != 13) {
    std::cerr << "Error in SW" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(3,3),lti::ipoint(5,5)) != 16) {
    std::cerr << "Error in SE" << endl;
    res = false;
  }

  // the sides
  if (integrator.sum(ii,lti::ipoint(0,-2),lti::ipoint(2,0)) != 6) {
    std::cerr << "Error in N+0" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(1,-2),lti::ipoint(3,0)) != 9) {
    std::cerr << "Error in N+1" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(0,3),lti::ipoint(2,5)) != 42) {
    std::cerr << "Error in S+0" << endl;
    res = false;
  }
  if (integrator.sum(ii,lti::ipoint(1,3),lti::ipoint(3,5)) != 45) {
    std::cerr << "Error in S+1" << endl;
    res = false;
  }
  if (integrator.sum(ii,lti::ipoint(-2,0),lti::ipoint(0,2)) != 15) {
    std::cerr << "Error in L+0" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(-2,1),lti::ipoint(0,3)) != 27) {
    std::cerr << "Error in L+1" << endl;
    res = false;
  }
  if (integrator.sum(ii,lti::ipoint(3,0),lti::ipoint(5,2)) != 24) {
    std::cerr << "Error in R+0" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(3,1),lti::ipoint(5,3)) != 36) {
    std::cerr << "Error in R+1" << endl;
    res = false;
  }

  // center
  if (integrator.sum(ii,lti::ipoint(0,0),lti::ipoint(2,2)) != 54) {
    std::cerr << "Error in Center 0" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(1,1),lti::ipoint(3,3)) != 99) {
    std::cerr << "Error in Center 1" << endl;
    res = false;
  }
  
  return res;
}


/*
 * Check Constant boundary
 */

bool checkConstant(const lti::matrix<lti::int32>& ii,
                   const lti::integralImage& integrator) {

  bool res = true;
  // Input image was:
  //
  //  1  2  3  4
  //  5  6  7  8
  //  9 10 11 12
  // 13 14 15 16

  // the corners
  
  if (integrator.sum(ii,lti::ipoint(-2,-2),lti::ipoint(0,0)) != 9) {
    std::cerr << "Error in NW" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(3,-2),lti::ipoint(5,0)) != 36) {
    std::cerr << "Error in NE" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(-2,3),lti::ipoint(0,5)) != 117) {
    std::cerr << "Error in SW" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(3,3),lti::ipoint(5,5)) != 144) {
    std::cerr << "Error in SE" << endl;
    res = false;
  }

  // the sides
  if (integrator.sum(ii,lti::ipoint(0,-2),lti::ipoint(2,0)) != 18) {
    std::cerr << "Error in N+0" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(1,-2),lti::ipoint(3,0)) != 27) {
    std::cerr << "Error in N+1" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(0,3),lti::ipoint(2,5)) != 126) {
    std::cerr << "Error in S+0" << endl;
    res = false;
  }
  if (integrator.sum(ii,lti::ipoint(1,3),lti::ipoint(3,5)) != 135) {
    std::cerr << "Error in S+1" << endl;
    res = false;
  }
  if (integrator.sum(ii,lti::ipoint(-2,0),lti::ipoint(0,2)) != 45) {
    std::cerr << "Error in L+0" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(-2,1),lti::ipoint(0,3)) != 81) {
    std::cerr << "Error in L+1" << endl;
    res = false;
  }
  if (integrator.sum(ii,lti::ipoint(3,0),lti::ipoint(5,2)) != 72) {
    std::cerr << "Error in R+0" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(3,1),lti::ipoint(5,3)) != 108) {
    std::cerr << "Error in R+1" << endl;
    res = false;
  }

  // center
  if (integrator.sum(ii,lti::ipoint(0,0),lti::ipoint(2,2)) != 54) {
    std::cerr << "Error in Center 0" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(1,1),lti::ipoint(3,3)) != 99) {
    std::cerr << "Error in Center 1" << endl;
    res = false;
  }
  
  return res;
}

/*
 * Check Periodic boundary
 */

bool checkPeriodic(const lti::matrix<lti::int32>& ii,
                   const lti::integralImage& integrator) {

  bool res = true;
  // Input image was:
  //
  //  1  2  3  4
  //  5  6  7  8
  //  9 10 11 12
  // 13 14 15 16

  // the corners
  
  if (integrator.sum(ii,lti::ipoint(-2,-2),lti::ipoint(0,0)) != 84) {
    std::cerr << "Error in NW" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(3,-2),lti::ipoint(5,0)) != 81) {
    std::cerr << "Error in NE" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(-2,3),lti::ipoint(0,5)) != 72) {
    std::cerr << "Error in SW" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(3,3),lti::ipoint(5,5)) != 69) {
    std::cerr << "Error in SE" << endl;
    res = false;
  }

  // the sides
  if (integrator.sum(ii,lti::ipoint(0,-2),lti::ipoint(2,0)) != 78) {
    std::cerr << "Error in N+0" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(1,-2),lti::ipoint(3,0)) != 87) {
    std::cerr << "Error in N+1" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(0,3),lti::ipoint(2,5)) != 66) {
    std::cerr << "Error in S+0" << endl;
    res = false;
  }
  if (integrator.sum(ii,lti::ipoint(1,3),lti::ipoint(3,5)) != 75) {
    std::cerr << "Error in S+1" << endl;
    res = false;
  }
  if (integrator.sum(ii,lti::ipoint(-2,0),lti::ipoint(0,2)) != 60) {
    std::cerr << "Error in L+0" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(-2,1),lti::ipoint(0,3)) != 96) {
    std::cerr << "Error in L+1" << endl;
    res = false;
  }
  if (integrator.sum(ii,lti::ipoint(3,0),lti::ipoint(5,2)) != 57) {
    std::cerr << "Error in R+0" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(3,1),lti::ipoint(5,3)) != 93) {
    std::cerr << "Error in R+1" << endl;
    res = false;
  }

  // center
  if (integrator.sum(ii,lti::ipoint(0,0),lti::ipoint(2,2)) != 54) {
    std::cerr << "Error in Center 0" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(1,1),lti::ipoint(3,3)) != 99) {
    std::cerr << "Error in Center 1" << endl;
    res = false;
  }
  
  return res;
}

/*
 * Check mirror boundary
 */

bool checkMirror(const lti::matrix<lti::int32>& ii,
                 const lti::integralImage& integrator) {

  bool res = true;
  // Input image was:
  //
  //  1  2  3  4
  //  5  6  7  8
  //  9 10 11 12
  // 13 14 15 16

  // the corners
  
  if (integrator.sum(ii,lti::ipoint(-2,-2),lti::ipoint(0,0)) != 24) {
    std::cerr << "Error in NW" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(3,-2),lti::ipoint(5,0)) != 45) {
    std::cerr << "Error in NE" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(-2,3),lti::ipoint(0,5)) != 108) {
    std::cerr << "Error in SW" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(3,3),lti::ipoint(5,5)) != 129) {
    std::cerr << "Error in SE" << endl;
    res = false;
  }

  // the sides
  if (integrator.sum(ii,lti::ipoint(0,-2),lti::ipoint(2,0)) != 30) {
    std::cerr << "Error in N+0" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(1,-2),lti::ipoint(3,0)) != 39) {
    std::cerr << "Error in N+1" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(0,3),lti::ipoint(2,5)) != 114) {
    std::cerr << "Error in S+0" << endl;
    res = false;
  }
  if (integrator.sum(ii,lti::ipoint(1,3),lti::ipoint(3,5)) != 123) {
    std::cerr << "Error in S+1" << endl;
    res = false;
  }
  if (integrator.sum(ii,lti::ipoint(-2,0),lti::ipoint(0,2)) != 48) {
    std::cerr << "Error in L+0" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(-2,1),lti::ipoint(0,3)) != 84) {
    std::cerr << "Error in L+1" << endl;
    res = false;
  }
  if (integrator.sum(ii,lti::ipoint(3,0),lti::ipoint(5,2)) != 69) {
    std::cerr << "Error in R+0" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(3,1),lti::ipoint(5,3)) != 105) {
    std::cerr << "Error in R+1" << endl;
    res = false;
  }

  // center
  if (integrator.sum(ii,lti::ipoint(0,0),lti::ipoint(2,2)) != 54) {
    std::cerr << "Error in Center 0" << endl;
    res = false;
  } 
  if (integrator.sum(ii,lti::ipoint(1,1),lti::ipoint(3,3)) != 99) {
    std::cerr << "Error in Center 1" << endl;
    res = false;
  }
  
  return res;
}


/*
 * A general validation method:
 * 1. Assume that the simpler zero boundary condition of the integral image
 *    class is ok.
 * 2. Extend the boundary of the test image to contain the proper boundary
 * 3. (Pseudo-) Randomly generate rectangles that cover part of the image 
 * 4. Compute the sum of elements in the extended image with zero boundary
 * 5. Compute the sum of elements with the proper boundary
 * 6. Compare the results
 */
bool check(const lti::integralImage& integrator,
           const lti::channel8& chnl8) {
  lti::integralImage refIntegrator(lti::Zero);
  
  const int offset = chnl8.rows()*4;

  // Expand the input channel
  lti::boundaryExpansion expander(offset,
                                  integrator.getParameters().boundaryType);

  lti::channel8 chnlE; // extended channel
  expander.apply(chnl8,chnlE);

  // The integral images
  lti::channel32 intImg,intImgRef;
  integrator.apply(chnl8,intImg);
  refIntegrator.apply(chnlE,intImgRef);

  // The random number generators
  lti::uniformDiscreteDistribution rngRow(0,chnl8.lastRow());
  rngRow.setSeed(1000);
  lti::uniformDiscreteDistribution rngCol(0,chnl8.lastColumn());
  rngCol.setSeed(2001);
  lti::uniformDiscreteDistribution rngSide(1-offset,offset-1);
  rngSide.setSeed(3002);

  const int numTests = 1000; 
  lti::irectangle wnd,ebWnd;

  int ok = 0;
  int failed = 0;

  int x,y;
  

  for (int i=0;i<numTests;++i) {
    y=rngRow.rand();
    x=rngCol.rand();
    
    // the window on the channel reference system
    wnd.setConsistent(x,y,x+rngSide.rand(),y+rngSide.rand());

    // the same window on the expanded channel reference system
    ebWnd.shift(wnd,lti::ipoint(offset,offset));
    
    const int val1 = integrator.sum(intImg,wnd);
    const int val2 = refIntegrator.sum(intImgRef,ebWnd);

    if (val1 == val2) {
      ++ok;
    } else {
      ++failed;
      std::cerr << wnd << " failed: result " << val1 << " and should be " 
                << val2 << endl;
    }
  }

  if (failed > 0) {
    std::cerr << ok << " tests ok." << endl;
    std::cerr << failed << " tests failed." << endl;
  }

  return (failed == 0);
}

/*
 * Main method
 */
int main(int argc, char* argv[]) {
  std::string file;    // image file
  int size;            // filter window size
  lti::eBoundaryType btype; // how to treat the boundary

  parseArgs(argc,argv,file,size,btype);
  
  lti::channel8 chnl8;  // a channel

  if (file.empty()) {
    // no image name given, let's create the test bench:
    //
    //  1  2  3  4
    //  5  6  7  8
    //  9 10 11 12
    // 13 14 15 16
    //
    chnl8.allocate(4,4);
    lti::channel8::iterator it = chnl8.begin();
    lti::channel8::iterator eit = chnl8.end();
    int i=1;
    while(it!=eit) {
      (*it++)=i++;
    }

    size = 3; // ensure the test-bed size of 3

  } else {
    // user indicated a file to be loaded
    lti::image img;       // the color image
    lti::ioImage loader;  // image loader
    if (!loader.load(file,img)) {
      // error reading the image, report why
      cerr << "Error loading image " << file << endl;
      cerr << loader.getStatusString() << endl;
      exit(EXIT_FAILURE);
    }

    chnl8.castFrom(img); // get a simple gray scale vesion of img
  }


  lti::channel32 intimg;// the integral image of channel
  lti::integralImage integrator(btype); // class to obtain an integral image
  
  // get the famous integral image
  integrator.apply(chnl8,intimg);

  if (!file.empty()) {
    // the user wants a filtered image
    
    // now, with the integral image compute a simple filter of the desired size
    const int hsize=size/2;
    const int norm =lti::sqr(2*hsize+1);
    lti::channel8 result(chnl8.size()); // an empty container for the result
    
    // just for fun, a simple time measurement
    lti::timer chronos;
    
    chronos.start();
    
    // this can be done much faster, but we ignore that there is an "internal"
    // region, where no boundary checks need to be done.
    for (int y=0;y<result.rows();++y) {
      for (int x=0;x<result.columns();++x) {
        result.at(y,x) 
          = integrator.sum(intimg,
                           // top-left coordinate of the window
                           lti::ipoint(x-hsize,y-hsize),
                           // bottom-right coordinate of the window
                           lti::ipoint(x+hsize,y+hsize))/norm;
      }
    }
    
    chronos.stop();
    
    cout << "Filter took " << chronos.getTime()/1000.0 << "ms. " << endl;
    
    viewer_type viewInt("Integral Image");
    viewer_type viewFil("Filtered Image");
    
    viewInt.show(intimg);
    viewFil.show(result);
    
    cout << "Press ENTER to continue" << endl;
    getchar();
  } else {
    // the test bed
    const lti::ubyte data[] = {2,3,5,7};
    lti::channel8 thinH(1,4,data);
    lti::channel8 thinV(4,1,data);
    lti::channel8 ptChnl(1,1,1);

    switch(btype) {
      case lti::Zero:
        cout << "Check for zero boundary: ";
        if (checkZero(intimg,integrator)) {
          cout << "ok" << endl;
        } else {
          cout << "failed" << endl;
        }
        break;
      
      case lti::Constant:
        cout << "Check for constant boundary: ";
        if (checkConstant(intimg,integrator) && 
            check(integrator,chnl8) && 
            check(integrator,thinH) && 
            check(integrator,thinV) &&
            check(integrator,ptChnl)) {
          cout << "ok" << endl;
        } else {
          cout << "failed" << endl;
        }

        
        break;

      case lti::Periodic:
        cout << "Check for periodic boundary: ";
        if (checkPeriodic(intimg,integrator) &&
            check(integrator,chnl8) && 
            check(integrator,thinH) && 
            check(integrator,thinV) &&
            check(integrator,ptChnl)) {
          cout << "ok" << endl;
        } else {
          cout << "failed" << endl;
        }
        break;

      case lti::Mirror:
        cout << "Check for mirror boundary: ";
        if (checkMirror(intimg,integrator) &&
            check(integrator,chnl8) && 
            check(integrator,thinH) && 
            check(integrator,thinV) &&
            check(integrator,ptChnl)) {
          cout << "ok" << endl;
        } else {
          cout << "failed" << endl;
        }
        break;

      default:
        cout << "No test bed for given case yet." << endl;
        break;
    }
  }


  return EXIT_SUCCESS;
}
