/*
 * Copyright (C) 1998-2004
 * Lehrstuhl fuer Technische Informatik, RWTH-Aachen, Germany
 * 
 * This file is part of the LTI-Computer Vision Library (LTI-Lib)
 *
 * The LTI-Lib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License (LGPL)
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * The LTI-Lib is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public 
 * License along with the LTI-Lib; see the file LICENSE.  If
 * not, write to the Free Software Foundation, Inc., 59 Temple Place -
 * Suite 330, Boston, MA 02111-1307, USA.  
 */ 

#include "ltiObject.h"
#include "ltiSplitImageTorgI.h"
#include "ltiIOImage.h"
#include "ltiDraw.h"
#include "ltiViewer2D.h"
#include "ltiMath.h"
#include "ltiPassiveWait.h"
#include "ltiTimer.h"

#include "ltiSurfLocalDescriptor.h"
#include "ltiFastHessianDetection.h"
#include "ltiLocationMatch.h"

#include "ltiMatrixTransform.h"
#include "ltiBilinearInterpolation.h"
#include "ltiSecondOrderStatistics.h"

#include "ltiLispStreamHandler.h"

#include <iostream>
#include <fstream>



int main(int argc, char* argv[]) {

  // get the image filename
  std::string file1("../../img/file1.png"),file2("../../img/file2.png");
  lti::image img1,img2;
  lti::channel chnl1,chnl2;
  lti::ioImage loader;
  
  lti::fastHessianDetection::parameters fhdPar;
  lti::surfLocalDescriptor::parameters surfPar;
  lti::locationMatch::parameters lmPar;
  
  const char* confFile = "match.dat";
  
  std::ifstream in(confFile);
  bool ok=false;
  if (in) {
    lti::lispStreamHandler lsh;
    lsh.use(in);
    ok =       lti::read(lsh,"file1",file1);
    ok = ok && lti::read(lsh,"file2",file2);
    ok = ok && lti::read(lsh,"fastHessianDetection",fhdPar);
    ok = ok && lti::read(lsh,"surfLocalDescriptor",surfPar);
    ok = ok && lti::read(lsh,"locationMatch",lmPar);
  }
  
  if (!ok) {
    std::ofstream out(confFile);
    lti::lispStreamHandler lsh;
    lsh.use(out);
    
    lti::write(lsh,"file1",file1);
    lti::write(lsh,"file2",file2);
    lti::write(lsh,"fastHessianDetection",fhdPar);
    lti::write(lsh,"surfLocalDescriptor",surfPar);
    lti::write(lsh,"locationMatch",lmPar);
  }
  
  lti::fastHessianDetection fhd(fhdPar);
  lti::surfLocalDescriptor surf(surfPar);
  lti::locationMatch lm(lmPar);
  
  // load a given pair of images
  loader.load(file1,img1);
  loader.load(file2,img2);
  
  if (img1.empty() || img2.empty()) {
    std::cerr << "Image(s) not found: " << file1 << " or " << file2 << "\n";
    std::cerr << "Please edit the file 'match.dat' and set there the images"
              << "\nyou want to use.\n";
    return 0;
  }
  
  chnl1.castFrom(img1);
  chnl2.castFrom(img2);
  
  // extract locations on both images
  lti::list<lti::location> locs1,locs2;
  fhd.apply(chnl1,locs1);
  fhd.apply(chnl2,locs2);
  
  std::cout << "Locs in img 1: " << locs1.size() << std::endl;
  std::cout << "Locs in img 2: " << locs2.size() << std::endl;
  
  std::cout << "Extracting descriptors...";
  std::cout.flush();
  
  // extract the descriptors for the locations
  std::list<lti::dvector> desc1,desc2;
  surf.apply(chnl1,locs1,desc1);
  surf.apply(chnl2,locs2,desc2);
  
  std::cout << " done." << std::endl;
  
  // find match
  std::cout << "Searching correspondences...";
  std::cout.flush();
  
  std::vector<lti::location> mlocs1,mlocs2;
  std::vector<lti::dvector> mdesc1,mdesc2;
  lti::timer chronos;
  chronos.start();
  lm.apply(locs1,desc1,locs2,desc2,mlocs1,mdesc1,mlocs2,mdesc2);
  chronos.stop();
  std::cout << " done. (" << chronos.getTime()/1000.0 << " ms)" << std::endl;
  
  std::cout << "Number of correspondences: " << mlocs1.size() << std::endl;
  
  // paint correspondences
  
  lti::draw<lti::rgbaPixel> painter;
  const int offset2 = 16+chnl1.columns();
  lti::image canvas(lti::max(chnl1.rows(),chnl2.rows()),
                    offset2 + chnl1.columns(),
                    lti::rgbaPixel(0,0,0,0));
  
  painter.use(canvas);
  painter.setColor(lti::rgbaPixel(128,220,255,128));
  
  canvas.fill(img1);
  canvas.fill(img2,0,16+chnl1.columns());
  
  for (unsigned int i=0;i<mlocs1.size();++i) {
    painter.line(lti::iround(mlocs1[i].position.x),
                 lti::iround(mlocs1[i].position.y),
                 lti::iround(mlocs2[i].position.x)+offset2,
                 lti::iround(mlocs2[i].position.y));
  }
  
  lti::viewer2D view1("Image 1");
  lti::viewer2D view2("Image 2");
  
  view1.show(chnl1);
  view2.show(chnl2);
  
  lti::viewer2D correspond("Correspondences");
  correspond.show(canvas);

  std::cout << "Press Enter to continue" << std::endl;
  getchar();
}

