/*
 * Copyright (C) 2009
 * Pablo Alvarado
 *
 * This file is part of the LTI-Computer Vision Library 2 (LTI-Lib-2)
 *
 * The LTI-Lib-2 is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the authors nor the names of its contributors may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/** 
 * \file   morphology.cpp
 *         Contains an simple example for the use of morphology classes
 * \author Pablo Alvarado
 *
 * \date   12.10.2009
 *
 * revisions ..: $Id: morphology.cpp,v 1.7 2012/07/21 03:10:46 alvarado Exp $
 */

#include <cstdlib>
#include <string>
#include <list>
#include <fstream>
#include <iostream>

#include <ltiErosion.h>
#include <ltiDilation.h>
#include <ltiDistanceTransform.h>
#include <ltiZhangSuenThinning.h>
#include <ltiSquareKernel.h>
#include <ltiOctagonalKernel.h>
#include <ltiEuclideanKernel.h>
#include <ltiCityBlockKernel.h>
#include <ltiIOImage.h>
#include <ltiViewer2D.h>
#include <ltiPassiveWait.h>
#include <ltiLispStreamHandler.h>

template class lti::squareKernel<float>;
template class lti::squareKernel<lti::ubyte>;

template class lti::octagonalKernel<float>;
template class lti::octagonalKernel<lti::ubyte>;

template class lti::euclideanKernel<float>;
template class lti::euclideanKernel<lti::ubyte>;

template class lti::cityBlockKernel<float>;
template class lti::cityBlockKernel<lti::ubyte>;

/**
 * Just a container for the example
 */
class example {
public:
  enum eOperation {
    Dilation,
    Erosion,
    Thinning,
    Opening,
    Closure,
    Distance
  };

  enum eKernel {
    Square,
    Octagonal,
    Euclidean,
    CityBlock
  };

  /**
   * Default constructor
   */
  example();

  /**
   * The execution method
   */
  int operator()(int argc,char *argv[]);

  /**
   * Read example configuration
   */
  bool read(lti::ioHandler& handler);

  /**
   * Write example configuration
   */
  bool write(lti::ioHandler& handler) const;
  
protected:
  /**
   * Default configuration file name
   */
  static const char *const defaultConfigFile_;

  /**
   * Configuration file name
   */
  std::string configurationFile_;

  /**
   * Threshold
   */
  float threshold_;

  /**
   * Binarization flag.
   *
   * Indicates if resulting image should be B/W or simply all values below
   * threshold should be set to 0 but the rest kept unchanged.
   */
  bool binarization_;

  /**
   * Erosion functor
   */
  lti::erosion erosion_;

  /**
   * Dilation functor
   */
  lti::dilation dilation_;

  /**
   * Thinning
   */
  lti::zhangSuenThinning thinning_;

  /**
   * Distance transform
   */
  lti::distanceTransform distance_;

  /**
   * Use dilation or erosion
   */
  eOperation operation_;

  /**
   * Kernel type
   */
  eKernel kernel_;

  /**
   * Size of Structuring Element
   */
  int seSize_;

  /**
   * Wait period
   */
  int wait_;

  /**
   * Print the usage of this example.
   */
  void usage(int argc,char *argv[]);

  /**
   * Parse the commands
   */
  bool parse(int argc,char *argv[],std::list<std::string>& files);
};

const char *const example::defaultConfigFile_ = "morphology.cfg";

example::example() {
  // initialization of the attributes
  configurationFile_  = defaultConfigFile_;
  threshold_=0.5f;
  binarization_=false;
  operation_ = Erosion;
  seSize_ = 15;
  wait_ = 10;
}

bool example::read(lti::ioHandler& handler) {
  bool b = true;

  b = lti::read(handler,"erosion",erosion_) && b;
  b = lti::read(handler,"dilation",dilation_) && b;
  b = lti::read(handler,"distance",distance_) && b;

  std::string str;
  
  b = lti::read(handler,"operation",str) && b;
  if (str == "Dilation") {
    operation_ = Dilation;
  } else if (str == "Erosion") {
    operation_ = Erosion;
  } else if (str == "Thinning") {
    operation_ = Thinning;
  } else if (str == "Opening") {
    operation_ = Thinning;
  } else if (str == "Closure") {
    operation_ = Thinning;
  } else {
    b = false;
  }
  b = lti::read(handler,"threshold",threshold_) && b;
  b = lti::read(handler,"binarization",binarization_) && b;
  b = lti::read(handler,"structuringElementSize",seSize_) && b;
  b = lti::read(handler,"wait",wait_) && b;

  return b;
}

bool example::write(lti::ioHandler& handler) const {
  bool b = true;

  b = lti::write(handler,"erosion",erosion_) && b;
  b = lti::write(handler,"dilation",dilation_) && b;
  b = lti::write(handler,"distance",distance_) && b;

  switch(operation_) {
  case Dilation: 
    b = lti::write(handler,"operation","Dilation") && b;
    break;
  case Erosion:
    b = lti::write(handler,"operation","Erosion") && b;
    break;
  case Thinning:
    b = lti::write(handler,"operation","Thinning") && b;
    break;
  case Opening:
    b = lti::write(handler,"operation","Opening") && b;
    break;
  case Closure:
    b = lti::write(handler,"operation","Closure") && b;
    break;
  default:
    b = lti::write(handler,"operation","Unknown") && b;
  }
    
  b = lti::write(handler,"threshold",threshold_) && b;
  b = lti::write(handler,"binarization",binarization_) && b;
  b = lti::write(handler,"structuringElementSize",seSize_) && b;
  b = lti::write(handler,"wait",wait_) && b;

  return b;
}

void example::usage(int argc,char *argv[]) {
  std::cout << "\nUsage: " << argv[0] << " [image] \n\n"; 
  std::cout << "  -t val Threshold in percentage (0-100)\n";
  std::cout << "  -b complete binarization\n";
  std::cout << "  -s val Size for structuring element (odd number 3-101)\n\n";

  std::cout << " Operations:\n";
  std::cout << "  -d dilate\n";
  std::cout << "  -e erode\n";
  std::cout << "  -i thinning\n\n";
  std::cout << "  -O opening\n";
  std::cout << "  -C closure\n\n";
  std::cout << "  -D distance\n\n";

  std::cout << " Kernels:\n";
  std::cout << "  -c city block\n";
  std::cout << "  -8 octagonal\n";
  std::cout << "  -2 square\n";
  std::cout << "  -E euclidean\n\n";

  std::cout << "  -w val wait pause between steps in ms\n";
  std::cout << "  -h Show this help\n" << std::endl;
}

bool example::parse(int argc,char *argv[],std::list<std::string>& files) {
  int i=1;

  files.clear();

  while (i<argc) {
    if ( (std::string(argv[i]) == "-h") ||
         (std::string(argv[i]) == "--help") ) {
      usage(argc,argv);
      return false;
    } else if ( (std::string(argv[i]) == "-t") ) {
      ++i;
      if (i<argc) {
        threshold_ = float(atoi(argv[i]))/100.0f;
      }
    } else if ( (std::string(argv[i]) == "-b") ) {
      binarization_=true;
    } else if ( (std::string(argv[i]) == "-w") ) {
      ++i;
      if (i<argc) {
        wait_ = atoi(argv[i]);
      }
    } else if ( (std::string(argv[i]) == "-s") ) {
      ++i;
      if (i<argc) {
        seSize_ = atoi(argv[i]);
        if (seSize_<3) {
          seSize_=3;
        } else if (seSize_%2 == 0) {
          seSize_++;
        }
      }
    } else if ( (std::string(argv[i]) == "-d") ) {
      operation_=Dilation;
    } else if ( (std::string(argv[i]) == "-e") ) {
      operation_=Erosion;
    } else if ( (std::string(argv[i]) == "-i") ) {
      operation_=Thinning;
    } else if ( (std::string(argv[i]) == "-O") ) {
      operation_=Opening;
    } else if ( (std::string(argv[i]) == "-C") ) {
      operation_=Closure;
    } else if ( (std::string(argv[i]) == "-D") ) {
      operation_=Distance;
    } else if ( (std::string(argv[i]) == "-c") ) {
      kernel_=CityBlock;
    } else if ( (std::string(argv[i]) == "-2") ) {
      kernel_=Square;
    } else if ( (std::string(argv[i]) == "-8") ) {
      kernel_=Octagonal;
    } else if ( (std::string(argv[i]) == "-E") ) {
      kernel_=Euclidean;
    } else {
      files.push_back(argv[i]);
    }
    ++i;
  }

  return true;
}

int example::operator()(int argc,char *argv[]) {

  std::list<std::string> argFiles;

  //
  // read the configuration file, and write it if something went wrong
  //
  std::ifstream in(configurationFile_.c_str());
  bool writeConfig = true;
  if (in) {
    lti::lispStreamHandler lsh(in);
    writeConfig = !read(lsh);
  }

  if (!parse(argc,argv,argFiles)) {
    return EXIT_FAILURE;
  }

  if (writeConfig) {
    std::ofstream out(configurationFile_.c_str());
    lti::lispStreamHandler lsh(out);
    if (!write(lsh)) {
      std::cerr << "Could not write file '" << configurationFile_ << "'" 
                << std::endl;

      return EXIT_FAILURE;
    }
    out << std::endl;
    out.close();
  }

  in.close();

  //
  // Now the real job!
  //

  lti::image img;

  // create or get an image

  if (argFiles.empty()) {
    // create an image
    img.assign(512,512,lti::rgbaPixel(0,0,0,0));
    img.fill(lti::rgbaPixel(255,255,255,255),64,64,511-64,511-64);
  } else {
    std::list<std::string>::const_iterator it = argFiles.begin();
    bool ok=true;
    lti::ioImage loader;
    // read the first valid image
    do {
      ok = loader.load(*it,img);
    } while (it!=argFiles.end() && !ok);
  }

  // convert to channel and do a simple thresholding
  lti::channel chnl;
  chnl.castFrom(img);

  lti::channel::iterator it,eit;
  eit=chnl.end();
  for (it=chnl.begin();it!=eit;++it) {
    if ((*it) < threshold_) {
      (*it)=0.0f;
    } else if (binarization_) {
      (*it)=1.0f;
    }
  }

  switch(kernel_) {
  case Euclidean: {
    lti::euclideanKernel<float> krn(seSize_);
    
    erosion_.setStructuringElement(krn);
    dilation_.setStructuringElement(krn);
    
  } break;
  case Octagonal: {
    lti::octagonalKernel<float> krn(seSize_);
    
    erosion_.setStructuringElement(krn);
    dilation_.setStructuringElement(krn);

  } break;
  case Square: {
    lti::squareKernel<float> krn(seSize_);
    
    erosion_.setStructuringElement(krn);
    dilation_.setStructuringElement(krn);

  } break;
  case CityBlock: {
    lti::cityBlockKernel<float> krn(seSize_);
    
    erosion_.setStructuringElement(krn);
    dilation_.setStructuringElement(krn);

  } break;
  default:
    std::cerr << "Unknown kernel type" << std::endl;
    return EXIT_FAILURE;
  }
  
  
  float oldSum,sum;

  // erode or dilate until the end!
  std::string title;
  switch(operation_) {
  case Dilation:
    title="Dilation";
    break;
  case Erosion:
    title="Erosion";
    break;
  case Thinning:
    title="Thinning";
    break;
  case Opening:
    title="Opening";
    break;
  case Closure:
    title="Closure";
    break;
  case Distance:
    title="Distance";
    break;
  default:
    title="Error";
    break;
  }


  lti::viewer2D viewe(title);
  sum = chnl.computeSumOfElements();
  do {
    viewe.show(chnl);
    if (wait_>0) {
      lti::passiveWait(wait_*1000);
    }
    switch(operation_) {
    case Dilation:
      dilation_.apply(chnl);
      break;
    case Erosion:
      erosion_.apply(chnl);
      break;
    case Thinning:
      thinning_.apply(chnl);
      break;
    case Opening:
      erosion_.apply(chnl);
      dilation_.apply(chnl);
      break;
    case Closure:
      dilation_.apply(chnl);
      erosion_.apply(chnl);
      break;
    case Distance:
      distance_.apply(chnl);
      break;
    default:
      std::cerr << "Operation unknown" << std::endl;
      return EXIT_FAILURE;
    }

    std::cout << ".";
    std::cout.flush();
    oldSum = sum;
    sum = chnl.computeSumOfElements();
  } while( sum != oldSum);

  std::cout << "\nClose the viewer to end." << std::endl;
  lti::viewer2D::interaction action;
  lti::ipoint pt;
  do {
    viewe.waitInteraction(action,pt);
  } while (action.action!= lti::viewer2D::Closed);
  

  return EXIT_SUCCESS;
}


int main(int argc,char *argv[]) {
  try {
    example obj;
    return obj(argc,argv);
  }
  catch (lti::exception& exp) {
    std::cout << "An LTI::EXCEPTION was thrown: ";
    std::cout << exp.what() << std::endl;
  }
  catch (std::exception& exp) {
    std::cout << "std::exception was thrown: ";
    std::cout << exp.what() << std::endl;
  }
  catch (...) {
    std::cout << "Unknown exception thrown!" << std::endl;
  }
  return EXIT_FAILURE;

} // main
