/*
 * Copyright (C) 2009,2010 by Pablo Alvarado
 * 
 * This file is part of the lecture CE-5201 Digital Image Processing and
 * Analysis, at the Costa Rica Institute of Technology.
 *
 * The LTI-Lib-2 is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the authors nor the names of its contributors may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "database.h"
#include "descriptor.h"
#include "progress.h"

#include <ltiLoadImageList.h>
#include <ltiTimer.h>
#include <ltiRound.h>
#include <ltiLispStreamHandler.h>

#include <iostream> // some output desired
#include <fstream> // some output desired
#include <string>
#include <list>


/*
 * Create an empty record
 */
database::record::record() 
  : lti::ioObject(),id(-1),roi(0,0,0,0),averageY(0.0f),filename("") {
}

/*
 * Create a record with the given data
 *
 * @param r ROI
 * @param y Average Y value of the ROI
 * @param f Filename of the image with the above data
 */
database::record::record(const int i,
                         const lti::irectangle& r,
                         const float y,
                         const std::string& f) 
  : lti::ioObject(),id(i),roi(r),averageY(y),filename(f) {
  
}

/*
 * Create an empty record
 */
database::record::record(const record& other) 
  : lti::ioObject(other),
    id(other.id),
    roi(other.roi),
    averageY(other.averageY),
    filename(other.filename) {
}

/*
 * Create an empty record
 */
database::record::~record()  {
  filename.clear();
}

/*
 * Returns the name of this class
 */
const std::string& database::record::name() const {
  _LTI_RETURN_CLASS_NAME
}

/*
 * Returns a copy of this object
 */
database::record* database::record::clone() const {
  return new record(*this);
}

/*
 * Returns a copy of this object
 */
database::record* database::record::newInstance() const {
  return new record;
}

/*
 * Write the parameters in the given ioHandler
 *
 * @param handler the ioHandler to be used
 * @param complete if true (the default) the enclosing begin/end will
 *        be also written, otherwise only the data block will be written.
 * @return true if write was successful
 *
 * This method has to be overloaded.  If not it always returns false.
 */
bool database::record::write(lti::ioHandler& handler, 
                             const bool complete) const {
  bool b = true;
  if (complete) {
    b = handler.writeBegin();
  }

  b = b && 
    lti::write(handler,"id",id) &&
    lti::write(handler,"roi",roi) &&
    lti::write(handler,"averageY",averageY) &&
    lti::write(handler,"filename",filename);
  
  if (complete) {
    b = b && handler.writeEnd();
  }
  
  return b;
}

/*
 * Read the parameters from the given ioHandler
 *
 * @param handler the ioHandler to be used
 * @param complete if true (the default) the enclosing begin/end will
 *        be also written, otherwise only the data block will be written.
 * @return true if write was successful
 *
 * This method has to be overloaded.  If not it always returns false.
 */
bool database::record::read(lti::ioHandler& handler,const bool complete) {
  bool b = true;
  if (complete) {
    b = handler.readBegin();
  }

  b = b && 
    lti::read(handler,"id",id) &&
    lti::read(handler,"roi",roi) &&
    lti::read(handler,"averageY",averageY) &&
    lti::read(handler,"filename",filename);
  
  if (complete) {
    b = b && handler.readEnd();
  }
  
  return b;
}

/*
 * Constructor
 */
database::database() {
}

/*
 * Destructor
 */
database::~database() {
  tree_.clear();
}

/*
 * Create the database from all images in the given directory
 */
bool database::create(const std::string& dir,
                      const descriptor& dip) {
  
  // load all images in a directory
  lti::loadImageList loader;             
  loader.useDirectory(dir,lti::loadImageList::Recurse);

  aspectRatio_ = dip.getAspectRatio();

  if (dip.getMaskLabels() <= 0) {
    std::cerr << "Mask not set yet" << std::endl;
    return false;
  }  
 
  std::cout << "Database creation for '" << dir << "' launched." << std::endl;


  lti::image img;

  // total number of images to be read
  const int grandTotal = loader.computeSize();
  progress info(grandTotal); // information about progress and estimated
                             // remaining time
  int counter(-1);

  info.start();

  while(loader.hasNext()) { // still images to be read in the directory

    counter++;
    lti::irectangle roi;
    float y;
    lti::fvector key;
    
    if (loader.load(img)){ // if image successfully loaded
      
      if (dip.describe(img,roi,y,key)) {
        record r(counter,roi,y,loader.getLastLoadedFilename());
        tree_.add(key,r);
      }

      // Show some time info
      info.print(std::cout,counter);
      std::cout << "\r";
      std::cout.flush();

    } // end of if image successfully loaded
      
  } // end of while

  info.stop(std::cout); // print how much time it took

  if (!tree_.build()) {
    std::cerr << "Error building kd-tree: " << tree_.getStatusString() 
              << std::endl;
    return false;
  }

  return true;
}

/*
 * Create the database from all images in the given directory
 */
bool database::query(const lti::fvector& key,
                     record& data) const {
  return tree_.searchNearest(key,data);
}

/*
 *
 */
bool database::query(const lti::fvector& key,
                     const int k,
                     tree_type::mmap_type& data) const {
  if (!tree_.searchNearest(k,key,data)) {
    std::cerr << "Error quering: " << tree_.getStatusString() << std::endl;
    return false;
  }
  return true;
}

/*
 * Save the database on the given filename
 */
bool database::save(const std::string& dbName) const {
  std::cout << "Saving databank in " << dbName << "..."; 
  std::cout.flush();

  std::ofstream out(dbName.c_str());
  if (out) {
    lti::lispStreamHandler lsh(out);

    if (lti::write(lsh,"aspectRatio",aspectRatio_) &&
        tree_.write(lsh)) {
      std::cout << "done." << std::endl; 
    } else {
      std::cout << "failed." << std::endl;
      return false;
    }
    out << std::endl;
    out.close();
  } else {
    std::cout << "failed." << std::endl;
    return false;
  }
  
  return true;
}

/*
 * Read the database from the given filename
 */
bool database::load(const std::string& dbName) {
  std::ifstream in(dbName.c_str());
  if (in) {
    lti::lispStreamHandler lsh(in);
    std::cout << "Loading databank '" << dbName << "' ..."; 
    std::cout.flush();

    if (lti::read(lsh,"aspectRatio",aspectRatio_) &&
        tree_.read(lsh)) {
      std::cout << "done." << std::endl; 
    } else {
      std::cout << "failed." << std::endl; 
      return false;
    }
    in.close();
  } else {
    return false;
  }
  return true;
}

