/*
 * Copyright (C) 2009 by Pablo Alvarado
 * 
 * This file is part of the lecture CE-5201 Digital Image Processing and
 * Analysis, at the Costa Rica Institute of Technology.
 *
 * The LTI-Lib-2 is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the authors nor the names of its contributors may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _DATABASE_H_
#define _DATABASE_H_

#include <ltiVector.h>
#include <ltiIoObject.h>
#include <ltiKdTree.h>
#include <ltiRectangle.h>

#include <string>

class descriptor;

/**
 * Class to manage the database 
 */
class database {
public:
  /**
   * Data stored in the database
   */
  class record : public lti::ioObject {
  public:
    /**
     * Create an empty record
     */
    record();

    /**
     * Create a record with the given data
     *
     * @param r ROI
     * @param y Average Y value of the ROI
     * @param f Filename of the image with the above data
     */
    record(const int id,
           const lti::irectangle& r,
           const float y,
           const std::string& f);

    /**
     * Copy constructor
     */
    record(const record& other);
           
    /**
     * Destroy record
     */
    virtual ~record();

    /**
     * ID of this record
     */
    int id;

    /**
     * Which part of the image has been described (Region Of Interest)
     */
    lti::irectangle roi;

    /**
     * Average value of the ROI in the Y channel of the Yuv color space
     */
    float averageY;

    /**
     * Complete path and filename to the image stored in the database
     */
    std::string filename;

    /**
     * Returns the name of this class
     */
    virtual const std::string& name() const;

    /**
     * Returns a copy of this object
     */
    virtual record* clone() const;

    /**
     * Returns a copy of this object
     */
    virtual record* newInstance() const;

    /**
     * Write the parameters in the given ioHandler
     *
     * @param handler the ioHandler to be used
     * @param complete if true (the default) the enclosing begin/end will
     *        be also written, otherwise only the data block will be written.
     * @return true if write was successful
     *
     * This method has to be overloaded.  If not it always returns false.
     */
    virtual bool write(lti::ioHandler& handler, const bool complete=true) const;

    /**
     * Read the parameters from the given ioHandler
     *
     * @param handler the ioHandler to be used
     * @param complete if true (the default) the enclosing begin/end will
     *        be also written, otherwise only the data block will be written.
     * @return true if write was successful
     *
     * This method has to be overloaded.  If not it always returns false.
     */
    virtual bool read(lti::ioHandler& handler,const bool complete=true);

  };

  /**
   * Type used for the kdtree
   */
  typedef lti::kdTree<lti::fvector,record> tree_type;

  /**
   * Constructor
   */
  database();

  /**
   * Destructor
   */
  ~database();

  /**
   * Create the database from all images in the given directory
   * @param dir Directory holding the images (it will be traversed recursively)
   * @param aspectRatio ensure the ROI will have the given aspect ratio
   */
  bool create(const std::string& dir,
              const descriptor& dip);

  /**
   * Search the nearest key in the database
   */
  bool query(const lti::fvector& key,
             record& data) const;

  /**
   * Search the k nearest elements to the key in the database
   */
  bool query(const lti::fvector& key,
             const int k,
             tree_type::mmap_type& data) const;

  /**
   * Save the database on the given filename
   */
  bool save(const std::string& dbName) const;

  /**
   * Read the database from the given filename
   */
  bool load(const std::string& dbName);

 private:

  /**
   * Kd-tree holding all data
   */
  tree_type tree_;

  /**
   * Remember the aspectRatio with which the database is created
   */
  float aspectRatio_;

  
};

#endif
