/*
 * Copyright (C) 2009 by Pablo Alvarado
 * 
 * This file is part of the lecture CE-5201 Digital Image Processing and
 * Analysis, at the Costa Rica Institute of Technology.
 *
 * The LTI-Lib-2 is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the authors nor the names of its contributors may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "descriptor.h"

#include <ltiVector.h>
#include <ltiRectangle.h>
#include <ltiIOImage.h>
#include <ltiDraw.h>
#include <ltiRound.h>
#include <ltiFastRelabeling.h>
#include <ltiSplitImageToYUV.h>
#include <ltiSplitImageToRGB.h>
#include <ltiMergeRGBToImage.h>
#include <ltiScaling.h>
#include <ltiConvolution.h>
#include <ltiGaussKernels.h>


#include <string>

/*
 * Constructor
 */
descriptor::descriptor() : aspectRatio_(4.0f/3.0f),labels_(0),antialias_(100) {
}

/*
 * Destructor
 */
descriptor::~descriptor() {
}

/*
 * Set the aspect ratio to be used
 */
void descriptor::setAspectRatio(const float aspectRatio) {
  aspectRatio_=aspectRatio; 
}

/*
 * Get the aspect ratio
 */
float descriptor::getAspectRatio() const {
  return aspectRatio_;
}

int descriptor::getMaskLabels() const {
  return labels_;
}

void descriptor::createMask() {
  const int side=256;
  const int hside=side/2;
  mask_.allocate(side,side);
  lti::draw<lti::ubyte> painter;
  painter.use(mask_);
  
  // Paint for rectangles covering the whole image, each one with its own label
  const bool filled = true;
  painter.setColor(1); painter.rectangle(0    ,0    ,hside-1,hside-1,filled);
  painter.setColor(2); painter.rectangle(hside,0    ,side-1 ,hside-1,filled);
  painter.setColor(3); painter.rectangle(0    ,hside,hside-1,side-1 ,filled);
  painter.setColor(4); painter.rectangle(hside,hside,side-1 ,side-1 ,filled);
  
  // Now paint a circle in the middle
  
  painter.setColor(lti::ubyte(0));
  
  // Ratio chosen such that all parts in the mask have the same area.
  painter.circle(hside-1,hside-1,lti::iround(side*0.2524f),filled);
  
  labels_ = 5;
  keyDim_ = labels_*3;
  
}  

bool descriptor::setMask(const std::string& maskFile) {
  lti::ioImage loader;
  lti::image img;

  if (loader.load(maskFile,img)) {
    lti::channel8 tmp;
    tmp.castFrom(img);
    lti::fastRelabeling::parameters frPar;
    frPar.sortSize = true;
    lti::fastRelabeling fr(frPar);

    fr.apply(tmp,mask_,labels_);

    keyDim_ = labels_*3;

  } else {
    createMask();
    loader.save(maskFile,mask_);
    return false; 
  }

  return true;
}  

/*
 * Compute the ROI and descriptor for that ROI of an entire image
 */
bool descriptor::describe(const lti::image& img,
                          lti::irectangle& roi,
                          float& y,
                          lti::fvector& key) const {

  // ---------------------
  // compute the valid ROI
  // ---------------------
  const int h = img.rows();
  const int w = img.columns();
  
  // try to keep width
  int height = lti::iround(float(w)/aspectRatio_);
  int width(w);
  if (height > h) {
    // didn't work, then keep height
    height=h;
    width =lti::iround(float(height)*aspectRatio_);

    int delta = (w-width)/2;
    roi.set(delta,0,delta+width-1,height-1);
  } else {
    // indeed, keep the width
    int delta = (h-height)/2;
    roi.set(0,delta,width-1,delta+height-1);
  }
  
  return computeKey(img,roi,y,key);
}

bool descriptor::computeKey(const lti::image& img,
                            const lti::irectangle& roi,
                                  float& y,
                                  lti::fvector& key) const {
  
  // Compute the mask steps
  float mx(0.0f),my(0.0f);
  const lti::ipoint roiSize(roi.getDimensions());
  const float dx = float(mask_.columns())/float(roiSize.x);
  const float dy = float(mask_.rows())/float(roiSize.y);

  key.assign(keyDim_,0.0f); // prepare to accumulate
  y=0;
  const int n=roiSize.x*roiSize.y;
  lti::ivector nLabel(labels_,0);
  float yy;
  int R,G,B,x,row; 

  static const float cr = 0.299f/255.0f;
  static const float cg = 0.587f/255.0f;
  static const float cb = 0.114f/255.0f;

  for (my=0.0f,row=roi.ul.y;row<=roi.br.y;++row,my+=dy) {
    for (mx=0.0f,x=roi.ul.x;x<=roi.br.x;++x,mx+=dx) {
      
      // read the pixel and compute its components
      const lti::rgbaPixel& px = img.at(row,x);
      px.get(R,G,B);
      yy=cr*R + cg*G + cb*B; // compute the Y value 
      y+=yy; // Accumulate Y value of the whole ROI
      
      // now compute which component in the descriptor we are
      // modifying
      const int idx = mask_.at(static_cast<int>(my),
                               static_cast<int>(mx));

      nLabel.at(idx)++;

      const int idx3 = 3*idx;

      key.at(idx3)  +=static_cast<float>(R);
      key.at(idx3+1)+=static_cast<float>(G);
      key.at(idx3+2)+=static_cast<float>(B);

    }
  }
  
  // now compute the corresponding averages
  y/=n;
  for (int i=0,i3=0;i<labels_;++i,i3+=3) {
    const int n = nLabel.at(i);
    if (n>0) {
      key.at(i3)  /=n;
      key.at(i3+1)/=n;
      key.at(i3+2)/=n;
    }
  }
  
  return true;
}
 
/*
 * Resize to tile size
 * @param imgFile string containing the file name of the image
 * @param bbox bounding box of the interesting region in of the image
 * @param tileSize size of the desired tile
 * @param tile the resulting tile image
 */
bool descriptor::resize(const std::string& imgFile,
                        const lti::irectangle& bbox,
                        const lti::ipoint& tileSize,
                        lti::image& tile) const {
  lti::ioImage loader;
  lti::image img;

  // read the image
  if (!loader.load(imgFile,img)) {
    std::cerr << "Error reading file '" << imgFile << "' : " 
              << loader.getStatusString() << std::endl;
    return false;
  }

  // Extract only the bounding box of the input image, to avoid
  // filtering unnecessary parts.
  lti::image subImg(img,bbox.ul.y,bbox.ul.x,bbox.br.y,bbox.br.x);

  // ensure proper size of destination
  tile.allocate(tileSize);

  // we are going to assume that the bbox and the tile have the same 
  // aspect ratio
  
  // compute scaling factor
  const float s = float(tileSize.x)/float(subImg.columns());

  // filter the image
  if (s < 1.0f) {
    // only necessary for downsampling

    // ensure and odd size at least as big as  window
    const int ksize = (lti::iround(ceil(100.0f/(s*antialias_)))/2)*2+1;

    if (ksize > 2) {
      // The kernel with auto variance
      lti::gaussKernel2D<float> gauss(ksize,-1);
      
      // Create and configure the convolution 
      lti::convolution::parameters cPar;
      cPar.boundaryType = lti::Constant;
      cPar.setKernel(gauss);
      lti::convolution filter(cPar);
      
      // Split to RGB
      lti::splitImageToRGB splitter;
      lti::channel r,g,b;
      
      // Split to the RGB components
      splitter.apply(subImg,r,g,b);
      
      // Low-pass filter channels to ensure Nyquist criterion when downsampling
      filter.apply(r);
      filter.apply(g);
      filter.apply(b);
      
      lti::mergeRGBToImage merger;
      merger.apply(r,g,b,subImg);
    }
  }
  
  // prepare for resizing to tile size
  lti::scaling<lti::rgbaPixel>::parameters scalePar;
  scalePar.boundaryType = lti::Constant;

  // we will force the tile to have the desired size
  scalePar.scale.set(float(tileSize.x)/float(subImg.columns()),
                     float(tileSize.y)/float(subImg.rows()));
  
  // create the scaler
  lti::scaling<lti::rgbaPixel> scaler(scalePar);

  scaler.apply(subImg,tile);

  return true;
}

void descriptor::setAntialias(const int aa) {
  antialias_ = aa;
}
