/*
 * Copyright (C) 2009 by Pablo Alvarado
 * 
 * This file is part of the lecture CE-5201 Digital Image Processing and
 * Analysis, at the Costa Rica Institute of Technology.
 *
 * The LTI-Lib-2 is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the authors nor the names of its contributors may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _DESCRIPTOR_H
#define _DESCRIPTOR_H

#include <ltiVector.h>
#include <ltiIoObject.h>
#include <ltiKdTree.h>
#include <ltiRectangle.h>
#include <ltiImage.h>
#include <ltiChannel8.h>

#include <string>

/**
 * Class to perform image processing tasks
 */
class descriptor {
public:

  /**
   * Constructor
   */
  descriptor();

  /**
   * Destructor
   */
  ~descriptor();

  /**
   * Set the aspect ratio to be used
   */
  void setAspectRatio(const float aspectRatio);

  /**
   * Get the aspect ratio
   */
  float getAspectRatio() const;

  /**
   * Read a mask from the given file, and analyze it.
   *
   * Return false if the mask could not be read from the file and a
   * new one needed to be created, which will be saved with the name
   * given.
   */
  bool setMask(const std::string& maskFile);

  /**
   * Get number of labels in the current set mask
   */
  int getMaskLabels() const;

  /**
   * Compute the ROI and descriptor for that ROI of an entire image
   */
  bool describe(const lti::image& img,
                lti::irectangle& roi,
                float& y,
                lti::fvector& key) const;

  /**
   * Compute the key for the given ROI
   */
  bool computeKey(const lti::image& img,
                  const lti::irectangle& roi,
                        float& y,
                        lti::fvector& key) const;

  /**
   * Resize to tile size
   * @param imgFile string containing the file name of the image
   * @param bbox bounding box of the interesting region in of the image
   * @param tileSize size of the desired tile
   * @param tile the resulting tile image
   */
  bool resize(const std::string& imgFile,
              const lti::irectangle& bbox,
              const lti::ipoint& tileSize,
              lti::image& tile) const;

  /**
   * Set antialias percentage
   */
  void setAntialias(const int aa);

 private:
  /**
   * Create a default mask
   */
  void createMask();

  /**
   * Aspect ratio
   */
  float aspectRatio_;

  /**
   * Mask being used
   */
  lti::channel8 mask_;

  /**
   * Labels in mask
   */
  int labels_;

  /**
   * Dimension of the key
   */
  int keyDim_;

  /**
   * Percentage of antialias
   */
  int antialias_;
  
};

#endif
