/*
 * Copyright (C) 2009 by Pablo Alvarado
 * 
 * This file is part of the lecture CE-5201 Digital Image Processing and
 * Analysis, at the Costa Rica Institute of Technology.
 *
 * The LTI-Lib-2 is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the authors nor the names of its contributors may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/** 
 * \file   mosaic.cpp
 *         First Lecture Project.
 * \author Pablo Alvarado
 * \date   06.08.2009
 * revisions ..: $Id: mosaic.cpp,v 1.1 2010/07/01 02:44:39 alvarado Exp $
 */

// LTI-Lib Headers
#include <ltiObject.h>
#include <ltiMath.h>     // General lti:: math and <cmath> functionality
#include <ltiIOImage.h>  
#include <ltiLispStreamHandler.h>
#include <ltiStatusMonitorCerr.h>
#include <ltiStatusMonitor.h>
#include <ltiRound.h>
#include <ltiChannel.h>
#include <ltiSplitImageToYUV.h>
#include <ltiMergeYUVToImage.h>
#include <ltiViewer2D.h>


#include "progress.h"   // To show progress

// Standard Headers: from ANSI C and GNU C Library
#include <cstdlib>  // Standard Library for C++
#include <getopt.h> // Functions to parse the command line arguments

// Standard Headers: STL
#include <iostream>
#include <string>
#include <map>
#include <ios>
#include <fstream>
#include <cstdlib>
#include <cstring>
#include <sys/stat.h>
#include <errno.h>
#include <dirent.h>

// Mosaic files
#include "mosaic.h"

// Ensure that the STL streaming is used.
using std::cout;
using std::cerr;
using std::endl;

// This directory on $HOME will contain databases and configurations
const char *const mosaic::dbDir = ".mosaic/";

// This file on $HOME will contain the configuration 
const char *const mosaic::configFile = "config";

// Get the $HOME value
std::string mosaic::getHome() const {
  std::string home = getenv("HOME");
  if (home[home.length()-1] != '/') {
    home+='/';
  }
  return home;
}

// Read configuration from file
bool mosaic::readConfig() {
  static const std::string file = 
    getHome()+std::string(dbDir)+std::string(configFile);

  std::cout << "Reading " << file << "...";
  std::cout.flush();

  std::ifstream in(file.c_str());
  if (in) {
    lti::lispStreamHandler lsh(in);
    if (lti::read(lsh,"aspectRatio",aspectRatio_)   &&
        lti::read(lsh,"maskName",maskName_)         &&
        lti::read(lsh,"outWidth",outWidth_)         && 
        lti::read(lsh,"tileWidth",tileWidth_)       && 
        lti::read(lsh,"numberHTiles",numberHTiles_) &&
        lti::read(lsh,"uniqRatio",uniqRatio_)       &&
        lti::read(lsh,"antialias",antialias_)       &&
        lti::read(lsh,"adaptY",adaptY_)             &&
        lti::read(lsh,"random",random_) ) {
      std::cout << "done." << std::endl;
      return true;
    }
  }

  std::cout << "failed." << std::endl;
  return false;
}

// Write configuration to file
bool mosaic::writeConfig() const {
  static const std::string dir = getHome()+std::string(dbDir);

  std::cout << "Check if '" << dir << "' exists";
  std::cout.flush();
  
  struct stat st;
  if (stat(dir.c_str(),&st) != 0) {
    // config directory does not exist.  Create it!
    if (mkdir(dir.c_str(),0777) != 0) {
      // Error
      std::cout << "\n  Error creating it: " << strerror(errno) << std::endl;
      return false;
    }
  }

  const std::string file = dir + std::string(configFile);

  std::cout << "\nCreating " << file << "...";
  std::cout.flush();

  std::ofstream out(file.c_str());
  if (out) {
    lti::lispStreamHandler lsh(out);
    if (lti::write(lsh,"aspectRatio",aspectRatio_)   &&
        lti::write(lsh,"maskName",maskName_)         &&
        lti::write(lsh,"outWidth",outWidth_)         && 
        lti::write(lsh,"tileWidth",tileWidth_)       && 
        lti::write(lsh,"numberHTiles",numberHTiles_) &&
        lti::write(lsh,"uniqRatio",uniqRatio_)       &&
        lti::write(lsh,"antialias",antialias_)       &&
        lti::write(lsh,"adaptY",adaptY_)             &&
        lti::write(lsh,"random",random_) ) {
        
      out << std::endl;
      std::cout << "done." << std::endl;
      return true;      
    }
  }

  std::cout << "failed." << std::endl;
  return false;
}

// Constructor
mosaic::mosaic(int argc, char* argv[]) 
  : task_(Usage),aspectRatio_(4.0f/3.0f),maskName_(""),
    outWidth_(-1),tileWidth_(32), numberHTiles_(128),uniqRatio_(1),
    antialias_(100),adaptY_(false),random_(false),verbose_(0) {

  // Default mask name
  maskName_ = getHome()+std::string(dbDir) + maskName_ + "mask.png";

  // Read default values from configuration file
  if (!readConfig()) {
    writeConfig();
  }
  
  // Read the desired mask
  dip_.setMask(maskName_);

  // Parse the command line
  parse(argc,argv);
}


/*
 * Help 
 */
void mosaic::usage() const {
  cout <<
    "usage: mosaic [options] <file>\n\n" \
    "   -c <dir> | --create <dir>   create an image database for " \
                                   "given directory\n" \
    "   -b <db>  | --database <db>  name of the database.  " \
                                   "If ommited while the\n" \
    "                               database is created, the name given \n" \
    "                               with -c is used.\n" \
    "   -l       | --list           list all available databases.\n" \
    "   -a <ar>  | --aspect <ar>    aspect ratio (4:3 or 1.33 notation)\n"\
    "   -w <w>   | --width <w>      width in pixels of the final image.\n" \
    "                               This is a suggestion to be adjusted\n" \
    "                               If negative, then the input image times\n"\
    "                               -w is used\n"\
    "   -t <w>   | --tile <w>       width of a tile in the final image\n" \
    "   -n <w>   | --number <w>     number of horizontal tiles in the final " \
                                   "image\n" \
    "   -u <u>   | --unique <u>     uniqueness ratio in L1 metric \n"\
    "   -f <%>   | --antialias <%>  Percentage of antialiasing filter \n" \
    "   -y                          Adjust brightness of each tile \n" \
    "   -v                          increase verbose level\n"\
    "   -r       | --random         Use random patterns\n"\
    "   -h       | --help           show this message.\n" \
    "   -o <img>                    name of the image file where the result \n"\
    "                               will be written.\n" \
    "   <file>                      input image\n\n" \
    "From the options -w -t and -n, you can give two in order to compute \n"\
    "the third one.\n" << std::endl;    
}

/*
 * Parse the line command arguments
 */
void mosaic::parse(int argc, char*argv[]) {

  int c;

  // We use the standard getopt.h functions here to parse the arguments.
  // Check the documentation of getopt.h for more information on this
  
  // structure for the long options. 
  static struct option lopts[] = {
    {"create"   ,required_argument,0,'c'},
    {"database" ,required_argument,0,'b'},
    {"aspect"   ,required_argument,0,'a'},
    {"width"    ,required_argument,0,'w'},    
    {"tile"     ,required_argument,0,'t'},
    {"number"   ,required_argument,0,'n'},
    {"unique"   ,required_argument,0,'u'},
    {"antialias",required_argument,0,'f'},
    {"list"     ,no_argument,0,'l'},
    {"random"   ,no_argument,0,'r'},
    {"help"     ,no_argument,0,'h'},
    {0,0,0,0}
  };

  int optionIdx;
  std::string sizeStack="tn";
  verbose_=0;
  while ((c = getopt_long(argc, argv, "c:b:o:n:t:w:a:u:f:rvylh",
                          lopts,&optionIdx)) != -1) {
    switch (c) {
    case 'c':
      task_=CreateDB;
      imagesDirectory_=optarg; // get the argument provided to -c
      break;
    case 'a':
      aspectRatio_ = parseAspect(optarg);
      std::cout << "Using aspect ratio " << aspectRatio_ << std::endl;
      break;
    case 'w':
      outWidth_ = atoi(optarg);
      sizeStack += 'w';
      break;
    case 't':
      tileWidth_ = atoi(optarg);
      sizeStack += 't';
      break;
    case 'u':
      uniqRatio_ = atoi(optarg);
      break;
    case 'f':
      antialias_ = atoi(optarg);
      break;
    case 'n':
      numberHTiles_ = atoi(optarg);
      sizeStack += 'n';
      break;
    case 'b':
      database_ = optarg;
      break;
    case 'o':
      outputImage_ = optarg;
      break;
    case 'l':
      listDataBases();
      task_ = Exit;
      return;
      break;
    case 'v':
      ++verbose_;
      break;
    case 'y':
      adaptY_ = true;
      break;
    case 'r':
      random_ = true;
      break;
    case 'h':
      break;
    default:
      cerr << "Option '-" << static_cast<char>(c) << "' not recognized." 
           << endl;
    }
  }

  // Now try to read the image name
  if (optind < argc) {             // if there are still some arguments left...
    if (task_!=CreateDB) {         // and we are not creating a database
      task_ = Mosaic;              // then we will do the photomosaic
      inputImage_ = argv[optind];  // with the given image file
    }
  }

  // at this point, we should have all, but we check if some automatic naming
  // is required
  if (database_.empty()) {    // if no database name was given
    if (task_ == CreateDB) {  // and we have to create the database
      // check if the imagesDirectory strind ends with a slash
      std::string::size_type pos =
        (imagesDirectory_[imagesDirectory_.length()-1] == '/') ?
        imagesDirectory_.rfind('/',imagesDirectory_.length()-2) :
        imagesDirectory_.rfind('/');
      
      if (pos == std::string::npos) {
        // no further slashes
        // just use the same directory name 
        database_ = imagesDirectory_;
      } else {
        database_ = imagesDirectory_.substr(pos+1);
      }
      if (database_[database_.length()-1]=='/') {
        database_.erase(database_.length()-1);
      }

      database_ = getHome()+std::string(dbDir) + database_ + ".db";

      std::cout << "Database name not specified.  Using '" 
                << database_ << "'" << std::endl;
    }
  } else if (task_ == Mosaic) {
    std::ifstream in(database_.c_str()); // try to open the file
    if (!in) { 
      // File not found, maybe is a normal database
      if (database_.rfind(".db") == std::string::npos) {
        database_+=".db";
      }

      if (database_.find("/") == std::string::npos) {
        database_= getHome()+std::string(dbDir) + database_;
      }
    }
    in.close();
  } else {
    // File not found, maybe is a normal database
    if (database_.rfind(".db") == std::string::npos) {
      database_+=".db";
    }
    
    if (database_.find("/") == std::string::npos) {
      database_= getHome()+std::string(dbDir) + database_;
    }
  }
  
  // if we have to do the mosaic and no output image name was given
  if ( (task_ == Mosaic) && (outputImage_.empty()) ) {
    
    if (inputImage_.empty()) {
      usage();
      task_ = Exit;
      return;
    }
    // start copying the name of the input image
    outputImage_ = inputImage_;

    // but find the last '.' to identify the extension
    std::string::size_type pos = outputImage_.rfind('.');
    // if we have an extension
    if (pos != std::string::npos) {
      // force to use png format, since jpeg looses too much on the smaller
      // images
      outputImage_.replace(pos,inputImage_.length(),"_mosaic.png"); 
    } else {
      outputImage_+="_mosaic.png";        // if not, just append postfix.png
    }
  }

  // now, mark what we have to later compute the parameters of size
  validVars_=sizeStack.substr(sizeStack.size()-2);
  if (validVars_[0]>validVars_[1]) {
    lti::swap(validVars_[0],validVars_[1]);
  }
}

float mosaic::parseAspect(const char *const str) const {
  std::string s(str);
  std::string::size_type pos;
  if ( (pos=s.find(':')) != std::string::npos) {
    int num = atoi(s.substr(0,pos-1).c_str());
    int den = atoi(s.substr(pos+1).c_str());
    return float(num)/den;
  } else {
    return atof(str);
  }
}

void mosaic::listDataBases() const {
  static const std::string stdDir =  getHome()+std::string(dbDir);
  
  std::string dbfile;
  std::string::size_type pos,width(0);
  DIR *dp;
  struct dirent *ep;
  dp = opendir (stdDir.c_str());
  if (dp == NULL) {
    std::cout << "\nNo databases found in '" << stdDir << "'" << std::endl;
    return;
  }

  std::map<std::string,std::string> files;
  
  while ( (ep = readdir(dp)) ) {
    dbfile = ep->d_name;
    if ((pos=dbfile.rfind(".db")) != std::string::npos) {
      std::string dbName = dbfile.substr(0,pos);
      files[dbName] = stdDir + dbfile;
      width = lti::max(width,dbName.size());
    }
  }

  width+=2;

  std::map<std::string,std::string>::const_iterator it,eit;

  std::cout << "\nDatabases found: \n" << std::endl;

  for (it=files.begin();it!=files.end();++it) {
    std::cout << " ";
    std::cout.width(width);
    std::cout << (*it).first << "  ->  " << (*it).second << std::endl;
  }
  
  std::cout << std::endl;

}


mosaic::eTasks mosaic::checkTask() const {
  return task_;
}

bool mosaic::createDB() {
  dip_.setAspectRatio(aspectRatio_);
  return (db_.create(imagesDirectory_,dip_) && db_.save(database_));
}

bool mosaic::createMosaic() {
  // Just a wrapper for a timer that pretty-prints the time
  progress totalInfo(1);
  totalInfo.start();

  // try to read the input image
  lti::ioImage loader;
  lti::image img;
  if (!loader.load(inputImage_,img)) {
    std::cerr << "Image '" << inputImage_ << "' could not be read\n"
              << "  Reported error was " << loader.getStatusString() 
              << std::endl;
    return false;
  }

  // read the database
  if (!db_.load(database_)) {
    std::cerr << "Database '" << database_ << "' could not be read\n"
              << std::endl;
    return false;
  }

  // compute the final image size 
  // three variables n t w
  if (outWidth_ < 0) {
    outWidth_ = img.columns()*abs(outWidth_);
  }

  if (validVars_ == "nt") {
    // compute the image width
    outWidth_ = numberHTiles_*tileWidth_;
  } else if (validVars_ == "nw") {
    // compute the tile width
    tileWidth_ = lti::iround(float(outWidth_)/numberHTiles_);
    // recompute the image width
    outWidth_ = numberHTiles_*tileWidth_;
  } else if (validVars_ == "tw") {
    // compute the number of tiles
    numberHTiles_ = lti::iround(float(outWidth_)/tileWidth_);
    // recompute the image width
    outWidth_ = numberHTiles_*tileWidth_;
  }

  // tile height on the output image
  const int tileHeight   = lti::iround(float(tileWidth_)/aspectRatio_);

  // tile width on the input image
  const float sTileWidth    = float(img.columns())/numberHTiles_;
  // tile height on the input image
  const float sTileHeight   = sTileWidth*tileHeight/tileWidth_;

  const int nvt = lti::iround(float(img.rows())/sTileHeight);
  
  const int numberVTiles = ((nvt*sTileHeight)>float(img.rows())) ? nvt-1 : nvt;
  const int outHeight    = numberVTiles*tileHeight;
  
  // Show some initial statistics on the work to be done
  std::cout << std::string(78,'-') << std::endl;

  std::cout << "Creating mosaic for '" << inputImage_ << "'" << std::endl;
  std::cout << "Input image size : " 
            << img.columns() << " x " << img.rows() << std::endl;
  std::cout << "Output image size: " 
            << outWidth_ << " x " << outHeight << std::endl;
  std::cout << "Tile size        : "
            << tileWidth_ << " x " << tileHeight 
            << "   \t Aspect ratio: " << aspectRatio_ << std::endl;
  std::cout << "Number of tiles  : "
            << numberHTiles_ << " x " << numberVTiles << std::endl;
  if (adaptY_) {
    std::cout << "Adaption of Y channel activated." << std::endl;
  }

  std::cout << std::string(78,'-') << std::endl;

  // Resize the output image
  lti::image outImage(outHeight,outWidth_);

  // total number of images to be read
  const int total=numberVTiles * numberHTiles_;
  progress info(total);
  int counter(0);

  // number of images required for the given uniqueness ratio
  // const int numImg = (lti::sqr(uniqRatio_*2+1)-1)/2 + 1;
  const int numImg = lti::sqr(uniqRatio_*2+1);

  // imatrix containing which image in the cache corresponds to a given
  // position
  lti::imatrix indices(numberVTiles,numberHTiles_,-1); // -1 means unassigned
  lti::channel yChannel(numberVTiles,numberHTiles_);

  // Avoid converting more than one time a specific image
  typedef std::map< int, std::pair< const database::record* ,
                                    lti::image > > cache_type;
  cache_type cache;

  // the matrix of records in the database
  simil_type simil;
  simil.allocate(numberVTiles,numberHTiles_);

  // Matrix of iterators
  itMatrix_type itMatrix(numberVTiles,numberHTiles_);

  // Pass the aspect ratio to the descriptor 
  dip_.setAspectRatio(aspectRatio_);

  // What were we doing?
  std::cout << "Searching for best images as tiles of the mosaic..." 
            << std::endl;

  // start counting time
  info.start();
  float fx(0.0f),fy(0.0f);
  lti::fvector key;
  
  if (not random_) {
    for (int ty=0;ty<numberVTiles;++ty) {
    
      const float ny=fy+sTileHeight;
      fx=0.0f;
      for (int tx=0;tx<numberHTiles_;++tx) {
      
        const float nx=fx+sTileWidth;
      
        // compute the descriptor and associated info
        const lti::irectangle roi(lti::iround(fx),lti::iround(fy),
                                  lti::iround(nx)-1,lti::iround(ny)-1);

        // get the descriptor for the tile
        dip_.computeKey(img,roi,yChannel.at(ty,tx),key);

        // Find most similar images and store them in a matrix
        if (!db_.query(key,numImg,simil.at(ty,tx))) {
          return false;
        }

        // Recursively try not to repeat the elements
        checkNeighborhood(ty,tx,uniqRatio_,simil,itMatrix,indices);

        // for the next iteration
        fx=nx;


        // ------------------------
        //   Show some time info 
        // ------------------------
        counter++;

        info.print(std::cout,counter);
        std::cout << "\r";
        std::cout.flush();

      }

      fy=ny;
                                                
    }
  } else { // if random then

    // the array of positions
    const int posSize = numberVTiles*numberHTiles_;
    lti::vector<lti::ipoint> pos(posSize);
    for (int ty=0,i=0;ty<numberVTiles;++ty) {
      for (int tx=0;tx<numberHTiles_;++tx) {
        pos.at(i++).set(tx,ty);
      }
    }

    // ensure a random analysis of the "pixels"
    scrambler_.apply(pos);
    
    for (int i=0;i<posSize;++i) {
      const int ty=pos.at(i).y;
      const int tx=pos.at(i).x;
      
      const float fy=ty*sTileHeight;
      const float ny=fy+sTileHeight;

      const float fx=tx*sTileWidth;
      const float nx=fx+sTileWidth;
      
      // compute the descriptor and associated info
      const lti::irectangle roi(lti::iround(fx),lti::iround(fy),
                                lti::iround(nx)-1,lti::iround(ny)-1);
      
      // get the descriptor for the tile
      dip_.computeKey(img,roi,yChannel.at(ty,tx),key);
      
      // Find most similar images and store them in a matrix
      if (!db_.query(key,numImg,simil.at(ty,tx))) {
        return false;
      }
      
      // Recursively try not to repeat the elements
      checkNeighborhood(ty,tx,uniqRatio_,simil,itMatrix,indices);
      
      // ------------------------
      //   Show some time info 
      // ------------------------
      info.print(std::cout,i);
      std::cout << "\r";
      std::cout.flush();
      
    } // for i

  } // end of else not random

  // This is the slowest part, specially because of the antialiasing filters!
  
  info.stop(std::cout); // print how much time it took
  std::cout << std::endl;
  const lti::ipoint tileSize(tileWidth_,tileHeight);

  if (verbose_ > 0) {
    static lti::viewer2D view("Intensity map of destination");
    view.show(yChannel);
  }
  

  std::cout << "Putting all tiles together... be patient" << std::endl;
  counter = 0;
  info.start();

  lti::splitImageToYUV yuvSplitter;
  lti::mergeYUVToImage yuvMerger;
  lti::channel y,u,v;

  // Now let's put the corresponding tiles
  // Now glue all individual tiles together.
  for (int ty=0,cy=0;ty<numberVTiles;++ty,cy+=tileHeight) {
    for (int tx=0,cx=0;tx<numberHTiles_;++tx,cx+=tileWidth_) {
      const lti::image* tile;
        
      // Try to find the given image index in the cache
      cache_type::iterator cit = cache.find(indices.at(ty,tx));
      if (cit != cache.end()) {
        // Image already here: just point to the already scaled tile
          
        const cache_type::mapped_type& cd = (*cit).second; // points to pair
        tile = &cd.second;
      } else {
        const mmap_type::const_iterator mit = itMatrix.at(ty,tx);
        // Get a reference to the 
        const database::record& rec = mit->second->data;
          
        // Add new element to the cache
        cit =
          cache.insert(std::make_pair(indices.at(ty,tx),
                                      std::make_pair(&rec,
                                                     lti::image()))).first;

        // Image not yet in cache... create it
        dip_.resize(rec.filename,rec.roi,tileSize,cit->second.second);
        tile = &(cit->second.second); // tile points to the new image in
        // cache
      }
        
      // Adjust intestities if the user wants that
      if (adaptY_) {
        // Get Y at the destination (what we want!)
        const float want = yChannel.at(ty,tx);
        const float have = (*cit).second.first->averageY;
          
        yuvSplitter.apply(*tile,y,u,v);
        y.multiply(want/have);
          
        lti::image tmpTile;
        yuvMerger.apply(y,u,v,tmpTile);
          
        // Insert image 
        outImage.fill(tmpTile,cy,cx);
      } else {
          
        // Insert image 
        outImage.fill(*tile,cy,cx);
      }
        
      // ------------------------
      //   Show some time info 
      // ------------------------
      counter++;
        
      info.print(std::cout,counter);
      std::cout << "\r";
      std::cout.flush();
    } // tx
  } // ty

  info.stop(std::cout); // print how much time it took
  std::cout << std::endl;
  std::cout << "Saving image '" << outputImage_ << "'" << std::endl;

  // Save the computed image
  if (!loader.save(outputImage_,outImage)) {
    std::cerr << "Image '" << outputImage_ << "' could not be written\n"
              << "  Reported error was " << loader.getStatusString() 
              << std::endl;
    return false;
  }

  totalInfo.stop(std::cout);

  std::cout << std::endl;

  if (verbose_ > 0) {
    static lti::viewer2D result("Mosaic");
    result.show(outImage);

    std::cout << "Close the window to finish..." << std::endl;
    lti::viewer2D::interaction action;

    do {
      lti::ipoint pos;
      result.waitInteraction(action,pos);
    } while(action.action != lti::viewer2D::Closed);
  }


  return true;
}

int mosaic::checkNeighborhood(const int row,
                              const int col,
                              const int radius,
                              const simil_type& simil,
                              itMatrix_type& itMatrix,
                              lti::imatrix& indices) const {

  // First: a complete valid range
  const int fromRow = lti::max(0,row-radius);
  const int toRow   = lti::min(indices.lastRow(),row+radius);
  const int fromCol = lti::max(0,col-radius);
  const int toCol   = lti::min(indices.lastColumn(),col+radius);
  
  // a reference to the current pixel results
  const mmap_type& best = simil.at(row,col);
  
  // check if this pixel has been assigned before:
  int tmpIdx = indices.at(row,col);

  mmap_type::const_iterator& it = itMatrix.at(row,col);
  const mmap_type::const_iterator eit = best.end();

  if (tmpIdx >= 0) {

    // maybe last time we were already at the end here
    if (it == eit) {
      // don't change a thing, or maybe things could get worse
      return -1; // stop recursion
    }
    ++it;
    if (it == eit) {
      // holly cow!  We got to the end of the list and we did not find
      //             a thing!  
      // we should get a random element, but by now just return the best one
      indices.at(row,col)  = best.begin()->second->data.id;
      itMatrix.at(row,col) = best.begin();

      std::cout << "\n:-(\n";
      return -1; // stop recursion
    }
  } else {
    it = best.begin();
  }

  // try from the best to the worse to check if the value is repeated 
  for (;it!=eit;++it) {
    // "it" points to the current best pixel value
    const int index = it->second->data.id; // index in the middle
    const int dist  = it->first;
    //
    bool abort = false;
    for (int y=fromRow;!abort and y<=toRow;++y) {
      for (int x=fromCol;!abort and x<=toCol;++x) {
        if ((y==row) and (x==col)) {
          continue;
        }
        const int oi = indices.at(y,x); // index on the neighborhood
        if (oi<0) {
          if (random_) {
            continue;
          } else {
            // seems we got into unexplored territory => we're done!
            return (indices.at(row,col) = index);
          }
        }

        if (index == oi) { 
          // oh no! the best image was already used
          
          // check if the distance is best here than there
          if (dist < itMatrix.at(y,x)->second->data.id) {
            // distance best here: recurse there
            checkNeighborhood(y,x,radius,simil,itMatrix,indices);
            // keep on with the next neighbors
          } else {
            // distance best there, so keep on with the next element
            abort = true;
          }
        } // end of if
      } // end of for x
    } // end of for y
    
    
    if (!abort) {
      // we have the best picture so far!
      return (indices.at(row,col) = index);
    }
    
  } // end of for it
  
  // well we tried every element similar to (row,col), and always was someone
  // better... so ignore everything and return the best so far
  
  indices.at(row,col)  = best.begin()->second->data.id;
  itMatrix.at(row,col) = best.begin();
  std::cout << "\n:-(\n";

  return -1; // we should not ever come here!
}



bool mosaic::apply() {

  lti::statusMonitorCerr smerr;
  lti::setStatusMonitor(smerr);

  switch(checkTask()) {
  case CreateDB:
    return createDB();
    break;
  case Mosaic:
    return createMosaic();
    break;
  case Exit:
    return true;
    break;
  case Usage:
    usage();
    return true;
    break;
  default:
    usage();
  }
  return true;
}

/*
 * Main method
 */
int main(int argc, char* argv[]) {
  
  mosaic myMosaic(argc,argv);
  if (myMosaic.apply()) {
    return EXIT_SUCCESS;
  }

  return EXIT_FAILURE;
    

}

