/*
 * Copyright (C) 2009,2010 by Pablo Alvarado
 * 
 * This file is part of the lecture CE-5201 Digital Image Processing and
 * Analysis, at the Costa Rica Institute of Technology.
 *
 * The LTI-Lib-2 is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the authors nor the names of its contributors may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _MOSAIC_H
#define _MOSAIC_H

#include <string>
#include <ltiMatrix.h>
#include <ltiScramble.h> 

#include "database.h"
#include "descriptor.h"
#include "minimatrix.h" // To hold a matrix of std::list or whatever

/**
 * This class is used to create a photo-mosaic
 */
class mosaic {
  /**
   * Tasks
   */
  enum eTasks {
    CreateDB, /*!< Create the image database */
    Mosaic,   /*!< Create the image mosaic from the database */
    Usage,    /*!< Exit showing the usage first */
    Exit      /*!< Exit the program */
  };

public:

  /**
   * Constructor with the command line parameters
   */
  mosaic(int argc, char*argv[]);

  /**
   * Print how to use the command line
   */
  void usage() const;

  /**
   * Check if we are ready and have to exit
   */
  eTasks checkTask() const;

  /**
   * Create the data base
   * \return true if successful, false otherwise
   */
  bool createDB();

  /**
   * Create the data base
   * \return true if successful, false otherwise
   */
  bool createMosaic();

  /**
   * Do the real job.
   * \return true if successful, false otherwise
   */
  bool apply();

private:

  /**
   * Parse the command line arguments
   */
  void parse(int argc, char*argv[]);

  /**
   * List available data bases
   */
  void listDataBases() const;

  /**
   * Read configuration
   */
  bool readConfig();

  /**
   * Write configuration
   */
  bool writeConfig() const;

  /**
   * Return the home directory of the user
   */
  std::string getHome() const;

  /**
   * Configuration file
   */
  static const char *const configFile;

  /**
   * Database directory
   */
  static const char *const dbDir;

  /**
   * Random number generator
   */
  mutable lti::scramble scrambler_;

  /**
   * Attributes
   */
  //@{
  /**
   * Name of the database file
   */
  std::string database_;

  /**
   * Name of the image directory
   */
  std::string imagesDirectory_;

  /**
   * Name of the resulting image
   */
  std::string outputImage_;

  /**
   * Name of the input image
   */
  std::string inputImage_;

  /**
   * Flag to exit the program
   */
  eTasks task_;

  /**
   * Aspect ratio
   */
  float aspectRatio_;

  /**
   * Name of the mask used
   */
  std::string maskName_;

  /**
   * Database manager
   */
  database db_;

  /**
   * Digital Image Processor
   */
  descriptor dip_;

  /**
   * Width in pixels of the final image.
   *
   * If negative, the absolute value means how many times the size of the
   * input image has to be used
   */
  int outWidth_;

  /**
   * Width in pixels of a tile
   */
  int tileWidth_;

  /**
   * Number of horizontal tiles in the final image
   */
  int numberHTiles_;

  /**
   * Uniqueness ratio
   *
   * On an L1 region of the given ratio (i.e. a square around a tile)
   * no images will be repeated
   */
  int uniqRatio_;

  /**
   * Percentage of antialias filter.
   */
  int antialias_;

  /**
   * Flag to indicate brightness adaption
   */
  bool adaptY_;

  /**
   * Random patterns
   */
  bool random_;

  /**
   * String holding which of the above three variables holds the right value
   * to compute the third one
   */
  std::string validVars_;

  /**
   * Verbose level
   */
  int verbose_;
  //@}

  /**
   * Parse a string 
   */
  float parseAspect(const char *const str) const;

  /**
   * Typedef for the multimap
   *
   * Type of multi maps having the data ordered by increasing distances to key
   * with the record* as second elements
   */
  typedef database::tree_type::mmap_type mmap_type;

  /**
   * Typedef for the most similar images record
   */
  typedef minimatrix<mmap_type> simil_type;

  /**
   * Typedef for the matrix of iterators
   */
  typedef minimatrix< mmap_type::const_iterator > itMatrix_type;
  
  /**
   * Recursively search for best match, with Edison and Melissa's algorithm.
   *
   * Return the image finally assigned to the given position, or -1 if it
   * was not possible to find a suitable image
   */
  int checkNeighborhood(const int row,
                        const int col,
                        const int radius,
                        const simil_type& simil,
                        itMatrix_type& itMatrix,
                        lti::imatrix& indices) const;
  
};

#endif
