/*
 * Copyright (C) 1998-2004
 * Lehrstuhl fuer Technische Informatik, RWTH-Aachen, Germany
 *
 * This file is backported from the Computer Vision and Robotics Library
 * (CVR-Lib ) for the LTI-Lib-2
 *
 * The CVR-Lib is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the authors nor the names of its contributors may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */


#include "ltiObject.h"
#include "ltiSplitImageTorgI.h"
#include "ltiIOImage.h"
#include "ltiDraw.h"
#include "ltiViewer2D.h"
#include "ltiMath.h"
#include "ltiPassiveWait.h"
#include "ltiTimer.h"

#include "ltiSurfLocalDescriptor.h"
#include "ltiFastHessianDetection.h"
#include "ltiLocationMatch.h"

#include "ltiMatrixTransform.h"
#include "ltiBilinearInterpolation.h"
#include "ltiSecondOrderStatistics.h"

#include "ltiMergeRGBToImage.h"

#include "ltiEuclideanTransformation2D.h"
#include "ltiSimilarityTransformation2D.h"
#include "ltiAffineTransformation2D.h"
#include "ltiRansacEstimation.h"

#include "ltiLispStreamHandler.h"

#include <iostream>
#include <fstream>

template class lti::ransacEstimation<lti::similarityTransformation2D>;

int main(int argc, char* argv[]) {


  typedef lti::similarityTransformation2D base_trans;
  typedef lti::ransacEstimation<base_trans> ransac_type;

  // get the image filename
  std::string file1("../../img/boats2.png"),file2("../../img/boats.png");
  lti::image img1,img2;
  lti::channel chnl1,chnl2,chnl2e;
  lti::ioImage loader;

  lti::fastHessianDetection::parameters fhdPar;
  lti::surfLocalDescriptor::parameters surfPar;
  lti::locationMatch::parameters lmPar;
  ransac_type::parameters rsPar;

  const char* confFile = "ransac.dat";

  std::ifstream in(confFile);
  bool ok=false;
  if (in) {
    lti::lispStreamHandler lsh;
    lsh.use(in);
    ok =       lti::read(lsh,"file1",file1);
    ok = ok && lti::read(lsh,"file2",file2);
    ok = ok && lti::read(lsh,"fastHessianDetection",fhdPar);
    ok = ok && lti::read(lsh,"surfLocalDescriptor",surfPar);
    ok = ok && lti::read(lsh,"locationMatch",lmPar);
    ok = ok && lti::read(lsh,"ransac",rsPar);
  }

  if (!ok) {
    std::ofstream out(confFile);
    lti::lispStreamHandler lsh;
    lsh.use(out);

    lti::write(lsh,"file1",file1);
    lti::write(lsh,"file2",file2);
    lti::write(lsh,"fastHessianDetection",fhdPar);
    lti::write(lsh,"surfLocalDescriptor",surfPar);
    lti::write(lsh,"locationMatch",lmPar);
    lti::write(lsh,"ransac",rsPar);
  }

  lti::fastHessianDetection fhd(fhdPar);
  lti::surfLocalDescriptor surf(surfPar);
  lti::locationMatch lm(lmPar);
  ransac_type rs(rsPar);

  // load a given pair of images
  loader.load(file1,img1);
  loader.load(file2,img2);

  if (img1.empty() || img2.empty()) {
    std::cerr << "Image(s) not found: " << file1 << " or " << file2 << "\n";
    std::cerr << "Please edit the file 'match.dat' and set there the images"
              << "\nyou want to use.\n";
    return 0;
  }

  chnl1.castFrom(img1);
  chnl2.castFrom(img2);

  // extract locations on both images
  lti::list<lti::location> locs1,locs2;
  fhd.apply(chnl1,locs1);
  fhd.apply(chnl2,locs2);

  std::cout << "Locs in img 1: " << locs1.size() << std::endl;
  std::cout << "Locs in img 2: " << locs2.size() << std::endl;

  std::cout << "Extracting descriptors...";
  std::cout.flush();

  // extract the descriptors for the locations
  std::list<lti::dvector> desc1,desc2;
  surf.apply(chnl1,locs1,desc1);
  surf.apply(chnl2,locs2,desc2);

  std::cout << " done." << std::endl;

  // find match
  std::cout << "Searching correspondences...";
  std::cout.flush();

  std::vector<lti::location> mlocs1,mlocs2;
  std::vector<lti::dvector> mdesc1,mdesc2;
  lti::timer chronos;
  chronos.start();
  lm.apply(locs1,desc1,locs2,desc2,mlocs1,mdesc1,mlocs2,mdesc2);
  chronos.stop();
  std::cout << " done. (" << chronos.getTime()/1000.0 << " ms)" << std::endl;

  std::cout << "Number of correspondences: " << mlocs1.size() << std::endl;

  // estimate transform

  base_trans::parameters stPar;
  std::vector<lti::fpoint> pts1(mlocs1.size()),pts2(mlocs2.size());
  for (unsigned int i=0;i<mlocs1.size();++i) {
    pts1[i].copy(mlocs1[i].position);
    pts2[i].copy(mlocs2[i].position);
  }

  std::cout << "\nEstimating transformation with RANSAC...";
  std::cout.flush();
  
  chronos.start();
  lti::ivector inliers;
  rs.apply(pts1,pts2,inliers,stPar);
  chronos.stop();
  std::cout << " done. (" << chronos.getTime()/1000.0 << " ms)" << std::endl;
  std::cout << "Inliers: " << inliers.size() << " out of " << pts1.size() 
            << std::endl;
  
  lti::lispStreamHandler lsh;
  lsh.use(std::cout);
  stPar.write(lsh);
  std::cout << std::endl;

  // paint correspondences

  lti::draw<lti::rgbaPixel> painter;
  const int offset2 = 16+chnl1.columns();
  lti::image canvas(lti::max(chnl1.rows(),chnl2.rows()),
                    offset2 + chnl1.columns(),
                    lti::rgbaPixel(0,0,0,0));

  painter.use(canvas);
  painter.setColor(lti::rgbaPixel(255,250,245,128));

  canvas.fill(img1);
  canvas.fill(img2,0,16+chnl1.columns());

  for (unsigned int i=0;i<mlocs1.size();++i) {
    painter.line(lti::iround(mlocs1[i].position.x),
                 lti::iround(mlocs1[i].position.y),
                 lti::iround(mlocs2[i].position.x)+offset2,
                 lti::iround(mlocs2[i].position.y));
  }

  lti::viewer2D view1("Image 1");
  lti::viewer2D view2("Image 2");

  view1.show(chnl1);
  view2.show(chnl2);

  lti::viewer2D correspond("Correspondences");
  correspond.show(canvas);
  
  // overlay two channels as green and magenta channels
  base_trans st(stPar);

  typedef lti::matrixTransform<float> trans_type;
  trans_type::parameters mtPar;

  // First a keep dimensions
  mtPar.resizeMode = trans_type::KeepDimensions;
  st.generateMatrix(mtPar.transformation);

  // create a matrix transform object to map one of the images
  trans_type mt(mtPar);

  // ensure the size of the channel is the same as the other one.
  const lti::ipoint largest(lti::max(chnl1.columns(),chnl2.columns()),
                            lti::max(chnl1.rows(),chnl2.rows()));
  chnl1.resize(largest,0.0f);
  chnl2.resize(largest,0.0f);

  mt.apply(chnl1,chnl2e);
  
  lti::mergeRGBToImage merger;
  merger.apply(chnl2e,chnl2,chnl2e,canvas);

  lti::viewer2D viewM("Merged");

  viewM.show(canvas);

  getchar();
  std::cout << "Press Enter to continue" << std::endl;
}

