/*
 * Copyright (C) 2008 by Pablo Alvarado
 * 
 * This file is part of the LTI-Computer Vision Library 2 (LTI-Lib-2)
 *
 * The LTI-Lib-2 is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the authors nor the names of its contributors may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

// LTI-Lib Headers
#include "ltiObject.h"
#include "ltiSammonMapping.h"
#include "ltiStreamProgressInfo.h"
#include "ltiLispStreamHandler.h"

#include "ltiScramble.h"
#include "ltiViewer2D.h"
#include "ltiDraw.h"
#include <iostream>
#include <fstream>
#include <vector>


#if HAVE_GTK
#include "ltiViewer2D.h" // The normal viewer
typedef lti::viewer2D viewer_type;
#else
#include "ltiExternViewer2D.h"  // An emergency exit to display images.
typedef lti::externViewer2D viewer_type;
#endif

// Standard Headers
#include <cstdlib>
#include <iostream>
#include <string>

using std::cout;
using std::cerr;
using std::endl;

/*
 * Help 
 */
void usage() {
  cout << "Usage: sammonMapping file1 [file2 [file3 ...]]" << endl;
  cout << "Computes the Sammon Mapping of the data given in the files into \n";
  cout << "2D. The files must contain the rows of the matrices enclosed in\n"
       << "parenthesis, as produced by sending to a std::stream an \n"
       << "lti::matrix"
       << std::endl;

}

/**
 * Reads a file of a matrix created with the operator<<.
 *
 * The format is very simple, each row is delimited by (x1 x2 x3... xn)
 */
bool readData(const std::string& filename,lti::matrix<double>& data,
              const int maxSize) {
  std::ifstream in(filename.c_str());
  lti::lispStreamHandler lsh;
  if (in) {
    lsh.use(in);
    
    // read the first line
    std::vector<std::vector<double> > mat;
    
    bool ok = true;
    double val;
    while (ok) {
      ok = lsh.tryBegin();
      if (ok) {
        mat.push_back(std::vector<double>());
        unsigned int i=mat.size()-1;
        std::vector<double>& row = mat[i];
        while(ok && !lsh.tryEnd()) {
          ok = ok && lsh.read(val);
          row.push_back(val);
        }
      }
    }
    
    if (!mat.empty()) {
      if (mat.size()> static_cast<unsigned int>(maxSize)) {
        // ups! the matrix is too huge, so let us take some random samples
        lti::ivector idx(mat.size());
        for (int i=0;i<idx.size();++i) {
          idx.at(i)=i;
        }

        lti::scramble scrambler;
        scrambler.apply(idx);

        // transfer just the first elements to data
        const unsigned int cols = mat[0].size();
        data.allocate(maxSize,cols);
        for (int i=0;i<data.rows();++i) {
          if (mat[i].size() == cols) {
            data.getRow(i).castFrom(mat[idx.at(i)]);
          } else {
            return false;
          }
        }
        
      } else {
        // transfer the data
        const unsigned int cols = mat[0].size();
        data.allocate(mat.size(),cols);
        for (int i=0;i<data.rows();++i) {
          if (mat[i].size() == cols) {
            data.getRow(i).castFrom(mat[i]);
          } else {
            return false;
          }
        }
      }
    }
    return true;
  }
  return false;
}


/*
 * Main method
 */
int main(int argc, char* argv[]) {
  
    
  static const char* file = "sammon.dat";

  lti::sammonMapping::parameters smPar;
 
  std::ifstream in(file);
  int canvasSize=800;
  int maxSetSize=1500;

  bool w=true;
  if (in) {
    lti::lispStreamHandler lsh;
    lsh.use(in);
    w=!read(lsh,"sammonMapping",smPar);
    w=w||!read(lsh,"canvasSize",canvasSize);
    w=w||!read(lsh,"maxSetSize",maxSetSize);
  }
  if (w) {
    std::ofstream out(file);
    lti::lispStreamHandler lsh;
    lsh.use(out);
    write(lsh,"sammonMapping",smPar);
    write(lsh,"canvasSize",canvasSize);
    write(lsh,"maxSetSize",maxSetSize);
  }
    
  lti::sammonMapping sm(smPar);

  // read matrices
  std::vector<lti::dmatrix> mats;

  for (int i=1;i<argc;++i) {
    lti::dmatrix mat;
      
    std::cout << "Reading " << argv[i] << "... ";
    std::cout.flush();
    if (readData(argv[i],mat,maxSetSize)) {
      mats.push_back(lti::dmatrix());
      mat.detach(mats[mats.size()-1]);
      std::cout << mats[mats.size()-1].size() << " done." << std::endl;
    } else {
      std::cout << "failed." << std::endl;
    }

  }

  if (mats.empty()) {
    return EXIT_FAILURE;
  }

  // create one huge matrix
  unsigned int totalRows=0;
  const int cols = mats[0].columns();

  for (unsigned int i=0;i<mats.size();++i) {
    totalRows+=mats[i].rows();
    if (mats[i].columns() != cols) {
      std::cerr << "Not all matrices have the same number of columns" 
                << std::endl;
      return EXIT_FAILURE;
    }
  }

  lti::dmatrix theMat(totalRows,cols);
  int r=0;
  for (unsigned int i=0;i<mats.size();++i) {
    theMat.fill(mats[i],r,0);
    r+=mats[i].rows();
  }

  // we want to see some info while training
  lti::streamProgressInfo prog(std::cout);
  sm.setProgressObject(prog);
    
  lti::dmatrix trans;
  if (!sm.apply(theMat,trans) || trans.empty()) {
    return EXIT_FAILURE;
  }

  lti::dmatrix bbox(2,trans.columns());
    
  // compute bounding box
  bbox.getRow(0).fill(trans.getRow(0));
  bbox.getRow(1).fill(trans.getRow(1));

  for (int i=1;i<trans.rows();++i) {
    bbox.getRow(0).apply(trans.getRow(i),lti::min);
    bbox.getRow(1).apply(trans.getRow(i),lti::max);
  }

  // paint the points in trans
  const int csize = canvasSize;
  const int border = 10;
  lti::image canvas(csize+2*border,
                    csize+2*border,
                    lti::rgbaPixel(255,255,255,0));

  double mx,bx,my,by;
  mx = csize/(bbox.at(1,0)-bbox.at(0,0));
  bx = -mx*bbox.at(0,0) + border;
    
  my = csize/(bbox.at(1,1)-bbox.at(0,1));
  by = -my*bbox.at(0,1) + border;

  static const lti::rgbaPixel pal[] = {0x000000FF, 0x0000FF00, 0x00FF0000,
                                       0x00FF00FF, 0x00FFFF00, 0x0000FFFF,
                                       0x00808080};
  static const int palSize = 7;

  static const char signs[] = "+ox*sdv<>";
  static const int signsSize = 9;

  lti::draw<lti::rgbaPixel> painter;
  painter.use(canvas);

  int idx=0;
  for (unsigned int j=0;j<mats.size();++j) {
    std::string style;
    style+= signs[j%signsSize];
    painter.setStyle(style.c_str());
    painter.setColor(pal[j%palSize]);
    for (int i=idx;i<idx+mats[j].rows();++i) {
      painter.marker(lti::iround(mx*trans.at(i,0)+bx),
                     lti::iround(my*trans.at(i,1)+by));
    }
    idx+=mats[j].rows();
  }

  viewer_type viewer("Reduction");
  viewer.show(canvas);
  getchar();

  return EXIT_SUCCESS;
}
