/*
 * Copyright (C) 2007 by Pablo Alvarado
 * 
 * This file is part of the LTI-Computer Vision Library 2 (LTI-Lib-2)
 *
 * The LTI-Lib-2 is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the authors nor the names of its contributors may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/** 
 * \file   surfLocalDescriptor.cpp
 *         Contains an example of use for the class lti::surfLocalDescriptor
 * \author Pablo Alvarado
 * \date   04.11.2007
 * revisions ..: $Id: surfLocalDescriptor.cpp,v 1.3 2010/02/09 15:58:08 alvarado Exp $
 */

// LTI-Lib Headers
#include "ltiObject.h"
#include "ltiSplitImageTorgI.h"
#include "ltiIOImage.h"
#include "ltiDraw.h"
#include "ltiMath.h"
#include "ltiPassiveWait.h"

#include "ltiSurfLocalDescriptor.h"

#include "ltiMatrixTransform.h"
#include "ltiBilinearInterpolation.h"
#include "ltiSecondOrderStatistics.h"

#include "ltiLispStreamHandler.h"

#if HAVE_GTK
#include "ltiViewer2D.h" // The normal viewer
typedef lti::viewer2D viewer_type;
#else
#include "ltiExternViewer2D.h"  // An emergency exit to display images.
typedef lti::externViewer2D viewer_type;
#endif

// Standard Headers
#include <cstdlib>
#include <iostream>
#include <string>
#include <fstream>

using std::cout;
using std::cerr;
using std::endl;

/*
 * Help 
 */
void usage() {
  cout << "Usage: surfLocalDescriptor [image] [-h]" << endl;
  cout << "Test the surfLocalDescritor class of the LTI-Lib.  It computes \n"
       << "the descriptor for several image rotations with the corresponding\n"
       << "rotated location and computes several statistics on that\n";
  cout << "An empty image name forces the creation of a gradient image\n" 
       << endl;
  cout << "The file \"surf.dat\" allows to configure the evaluated behaviour."
       << endl;
  cout << "  -h show this help." << endl;
}

/*
 * Parse the line command arguments
 */
void parseArgs(int argc, char*argv[], 
               std::string& filename) {
  
  filename.clear();
  // check each argument of the command line
  for (int i=1; i<argc; i++) {
    if (*argv[i] == '-') {
      switch (argv[i][1]) {
        case 'h':
          usage();
          exit(EXIT_SUCCESS);
          break;
        default:
          break;
      }
    } else {
      filename = argv[i]; // guess that this is the filename
    }
  }
}

/*
 * Main method
 */
int main(int argc, char* argv[]) {

  std::string imgFile;
  parseArgs(argc,argv,imgFile);

  static const char* confFile = "surf.dat";

  lti::image img;
  lti::channel chnl,chnl2;
  
  lti::ioImage loader; // used to load an image file

  viewer_type view("Original");

  // viewer to show vector results
#if HAVE_GTK  
  lti::viewer2D::parameters vpar;
  vpar.title = "Vectors";
  vpar.mappingType = lti::viewer2DPainter::Optimal;
  vpar.zoom = 4;
  lti::viewer2D viewvct(vpar);
#else
  viewer_type viewvct;
#endif
      
  // painter object used to paint the location
  lti::draw<lti::rgbaPixel> painter;
  painter.use(img);
  painter.setColor(lti::rgbaPixel(255,240,128));
  
  // matrixTransform is used to rotate the image
  typedef lti::matrixTransform< float > trans_type;
  
  trans_type::parameters transPar;
  transPar.resizeMode = lti::geometricTransformBase::KeepDimensions;
  transPar.interpolatorParams.boundaryType = lti::Constant;
  trans_type transformer(transPar);

  // load the configuration for the descriptor generator
  lti::surfLocalDescriptor::parameters surfPar;
    
  std::ifstream in(confFile);
  bool write=true;
  if (in) {
    lti::lispStreamHandler lsh;
    lsh.use(in);
    write=!surfPar.read(lsh);
  }
  if (write) {
    // something went wrong loading the data, so just write again to fix
    // the errors
    std::ofstream out(confFile);
    lti::lispStreamHandler lsh;
    lsh.use(out);
    surfPar.write(lsh);
  }
  
  // the real descriptor extractor
  lti::surfLocalDescriptor surf(surfPar);

  // load a given image or create a gradient
  if (imgFile.empty() || !loader.load(imgFile,img)) {
    img.allocate(256,256);
    for (int y=0;y<img.rows();++y) {
      img.getRow(y).fill(lti::rgbaPixel(y,y,y,0));
    }
  }

  // work with gray valued images
  chnl.castFrom(img);
    
  // create a fake location just to have a controlled environment
  lti::fpoint o(img.columns()/2.0f,img.rows()/2.0f);
  lti::location loc;
  loc.position.copy(o);
  loc.radius = 3.0f;
  
  // container for the results
  lti::fmatrix mat;
  lti::dmatrix vects(365/5,surfPar.signSplit ? 128 : 64); 
  
  // for computation of error statistics
  lti::secondOrderStatistics<double> stats;

  lti::dvector desc;
  lti::dvector variances,means;
  
  // test several orientations of the location
  for (int phase=0;phase<=360;phase+=30) {
      
    // test several orientations of the image
    for (int deg=0;deg<=360;deg+=5) {
      const float rad=lti::degToRad(static_cast<float>(deg));
      loc.angle = lti::degToRad(static_cast<float>(deg+phase));
        
      transformer.setMatrix(rotationMatrix(o,rad));
      transformer.apply(chnl,chnl2);
      
      surf.apply(chnl2,loc,desc);
      vects.getRow(deg/5).fill(desc);
      
      img.castFrom(chnl2);
      lti::location loc2(loc);
      loc2.radius*=10;
      painter.set(loc2,true);
      view.show(img);
    }
      
    stats.apply(vects,means,variances);
    double err = sqrt(variances.computeSumOfElements());
    std::cout << phase << " deg: Error: " << err << std::endl;

    lti::channel chnl3;
    chnl3.castFrom(vects);

    viewvct.show(chnl3);
  }

  std::cout << "Press Enter to continue" << std::endl;
  getchar();

  return EXIT_SUCCESS;
}
