/*
 * Copyright (C) 2011 by Pablo Alvarado
 * 
 * This file is part of the LTI-Computer Vision Library 2 (LTI-Lib-2)
 *
 * The LTI-Lib-2 is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the authors nor the names of its contributors may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/** 
 * \file   v4l2.cpp
 *         Simple test example using the v4l2 camera.
 * \author Pablo Alvarado
 * \date   31.05.2011
 * revisions ..: $Id: v4l2.cpp,v 1.3 2012/08/06 03:04:57 alvarado Exp $
 */

// LTI-Lib Headers
#include "ltiObject.h"
#include "ltiMath.h"     // General lti:: math and <cmath> functionality
#include "ltiConfig.h"   // To check if GTK is there
#include "ltiTimer.h"    // To measure time

#include "ltiChannel.h" // Monochromatic byte channels
#include "ltiImage.h"

#include "ltiConvolution.h"
#include "ltiGaussKernels.h"
#include "ltiLispStreamHandler.h"


#include "ltiV4l2.h"
#include "ltiViewer2D.h" // The normal viewer

// Standard Headers
#include <cstdlib>
#include <iostream>
#include <string>
#include <fstream>

using std::cout;
using std::cerr;
using std::endl;

bool color = false;

/*
 * Help 
 */
void usage() {
  cout << "Usage: v4l2 [options] [-h]" << endl;
  cout << "Capture images with the v4l2 camera and display them.\n";
  cout << "  -h show this help." << endl;
  cout << "  -c show original in color" << endl;
  cout << "Keys" << endl;
  cout << "  ? : Print this help\n";
  cout << "  - : Decrease space filter size";
  cout << "  + : Increase space filter size";
  cout << "  a : Decrease past time consideration";
  cout << "  s : Increase past time consideration";
}

/*
 * Parse the line command arguments
 */
void parseArgs(int argc, char*argv[]) {
  
  // check each argument of the command line
  for (int i=1; i<argc; i++) {
    if (*argv[i] == '-') {
      switch (argv[i][1]) {
      case 'h':
        usage();
        exit(EXIT_SUCCESS);
        break;
      case 'c':
        color = true;
        break;
      default:
        break;
      }
    } 
  }
}

/*
 * Main method
 */
int main(int argc, char* argv[]) {

  parseArgs(argc,argv);

  static const char* confFile = "v4l2.lsp";

  lti::v4l2::parameters param;
 
  // try to read the configuration file
  std::ifstream in(confFile);
  bool write=true;
  if (in) {
    lti::lispStreamHandler lsh;
    lsh.use(in);
    write=!param.read(lsh);
  }
  if (write) {
    // something went wrong reading, write a new configuration file
    std::ofstream out(confFile);
    lti::lispStreamHandler lsh;
    lsh.use(out);
    param.write(lsh);
  }
  
  // create the detector with the user specified configuration
  lti::v4l2 cam(param);

  lti::image img;
  lti::channel x,last,y,s;

  int kSize = 1; // spatial filter
  float coef = 0.0f; // temporal filter
  
  lti::gaussKernel2D<float> gk(kSize);
  lti::convolution::parameters cp;
  cp.setKernel(gk);
  cp.boundaryType = lti::Constant;
  lti::convolution filter(cp);

  lti::viewer2D::interaction action,lastAction;
  lti::viewer2D::parameters vPar;
  vPar.mappingType = lti::viewer2DPainter::Optimal;
  vPar.title = "Filtered";
  static lti::viewer2D view(vPar);
  static lti::viewer2D orig("Original");

  vPar.title = "Subtracted";
  vPar.mappingType = lti::viewer2DPainter::OffsetSlope;
  vPar.offset=0.5;
  vPar.slope =0.75;

  lti::viewer2D subs(vPar);
  lti::ipoint pos;
  bool ok,update(false);

  if (!cam.apply(img)) {
    cout << "Error: " << cam.getStatusString() << endl;
    exit(EXIT_FAILURE);
  }
  last.assign(img.size(),0.0f);


  do {
    ok = cam.apply(img);

    if (ok && color) {
      orig.show(img);
    }
    x.castFrom(img);
    filter.apply(x);
    
    // y(n) = k*x(n) + a y(n-1)
    
    y.addScaled(1.0f-coef,x,coef,last);
    s.subtract(x,y);

    subs.show(s);
    view.show(y);
    y.swap(last);

    view.getLastAction(action,pos);
    if (action.action == lti::viewer2D::KeyPressed)  {
      switch(action.key) {
      case 'x':
      case 'X':
        ok = false;
        break;
      case lti::viewer2D::RightKey:
      case '+':
        kSize = lti::min(kSize + 2,65);
        update = true;
        break;
      case lti::viewer2D::LeftKey:
      case '-':
        kSize = lti::max(1,kSize-2);
        update = true;
        break;
      case lti::viewer2D::UpKey:
      case 'a':
      case 'A':
        coef = lti::max(-1.f,coef-0.01f);
        cout << "Coefficient: " << coef << endl;
        break;
      case lti::viewer2D::DownKey:
      case 's':
      case 'S':
        coef = lti::min(1.0f,coef+0.01f);
        cout << "Coefficient: " << coef << endl;
        break;
      default:
        cout << "<" << action.key << ">" << std::endl;
      }

      if (update) {
        cout << "Kernel: " << kSize << endl;
        gk.generate(kSize,lti::sqr((kSize+2.0f)/6.0f));
        filter.setKernel(gk);
        update=false;
      }
    }
    lastAction = action;
  } while(ok);


  return EXIT_SUCCESS;
}
